from __future__ import annotations

from collections import defaultdict
from itertools import chain
from typing import Any

import numpy as np
import pandas as pd
from _collections import OrderedDict
from anndata import AnnData
from category_encoders import CountEncoder, HashingEncoder
from rich import print
from rich.progress import BarColumn, Progress
from sklearn.preprocessing import LabelEncoder, OneHotEncoder

from ehrapy import logging as logg
from ehrapy.anndata._constants import EHRAPY_TYPE_KEY, NON_NUMERIC_ENCODED_TAG, NON_NUMERIC_TAG, NUMERIC_TAG
from ehrapy.anndata.anndata_ext import _get_var_indices_for_type

multi_encoding_modes = {"hash"}
available_encodings = {"one-hot", "label", "count", *multi_encoding_modes}


def encode(
    data: AnnData,
    autodetect: bool | dict = False,
    encodings: dict[str, dict[str, list[str]]] | dict[str, list[str]] | str | None = "one-hot",
) -> AnnData:
    """Encode categoricals of an :class:`~anndata.AnnData` object.

    Categorical values could be either passed via parameters or are autodetected on the fly.
    The categorical values are also stored in obs and uns (for keeping the original, unencoded values).
    The current encoding modes for each variable are also stored in uns (`var_to_encoding` key).
    Variable names in var are updated according to the encoding modes used. A variable name starting with `ehrapycat_`
    indicates an encoded column (or part of it).

    Autodetect mode:
        This can be used for convenience and when there are many columns that need to be encoded.
        Note that missing values do not influence the result.
        By using this mode, every column that contains non-numerical values is encoded.
        In addition, every binary column will be encoded too.
        These are those columns which contain only 1's and 0's (could be either integers or floats).

    Available encodings are:
        1. one-hot (https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html)
        2. label (https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html)
        3. count (https://contrib.scikit-learn.org/category_encoders/count.html)
        4. hash (https://contrib.scikit-learn.org/category_encoders/hashing.html)

    Args:
        data: A :class:`~anndata.AnnData` object.
        autodetect: Whether to autodetect categorical values that will be encoded.
        encodings: Only needed if autodetect set to False.
                   A dict containing the encoding mode and categorical name for the respective column
                   or the specified encoding that will be applied to all columns.

    Returns:
        An :class:`~anndata.AnnData` object with the encoded values in X.

    Examples:
        >>> import ehrapy as ep
        >>> adata = ep.dt.mimic_2()
        >>> # encode all autodetected (non numerical) columns using label encoding
        >>> adata_encoded = ep.pp.encode(adata, autodetect=True)

        >>> # Example using autodetect with non-default encoding mode:
        >>> import ehrapy as ep
        >>> adata = ep.dt.mimic_2()
        >>> # encode all auto-detected (non numerical) columns using one hot encoding
        >>> adata_encoded = ep.pp.encode(adata, autodetect=True, encodings='one_hot_encoding')

        >>> # Example using custom encodings per columns:
        >>> import ehrapy as ep
        >>> adata = ep.dt.mimic_2()
        >>> # encode col1 and col2 using label encoding and encode col3 using one hot encoding
        >>> adata_encoded = ep.pp.encode(adata,
        >>>                              autodetect=False,
        >>>                              encodings={'label': ['col1', 'col2'], 'one-hot': ['col3']})
    """
    if isinstance(data, AnnData):
        # basic type checking for passed parameters when encoding a single AnnData object
        if isinstance(encodings, str) and not autodetect:
            raise ValueError("Passing a string for parameter encodings is only possible when using autodetect=True!")
        elif autodetect and not isinstance(encodings, (str, type(None))):
            raise ValueError(
                f"Setting encode mode with autodetect=True only works by passing a string (encode mode name) or None not {type(encodings)}!"
            )
        return _encode(adata=data, autodetect=autodetect, encodings=encodings)
    else:
        raise ValueError(f"Cannot encode object of type {type(data)}. Can only encode AnnData objects!")


def undo_encoding(
    data: AnnData,
    columns: str = "all",
) -> AnnData | None:
    """Undo the current encodings applied to all columns in X.

    This currently resets the AnnData object to its initial state.

    Args:
        data: The :class:`~anndata.AnnData` object
        columns: The names of the columns to reset encoding for. Defaults to all columns.

    Returns:
        A (partially) encoding reset :class:`~anndata.AnnData`

    Examples:
        >>> import ehrapy as ep
        >>> # adata_encoded is an encoded AnnData object
        >>> adata_undone = ep.pp.encode.undo_encoding(adata_encoded)
    """
    if isinstance(data, AnnData):
        return _undo_encoding(data, columns)
    else:
        raise ValueError(f"Cannot decode object of type {type(data)}. Can only decode AnnData objects!")


def _encode(
    adata: AnnData,
    autodetect: bool | dict = False,
    encodings: dict[str, dict[str, list[str]]] | dict[str, list[str]] | str = None,
) -> AnnData | None:
    """Encode the initial read AnnData object. Categorical values could be either passed via parameters or autodetected.

    Label encodes by default which is used to save initially unencoded AnnData objects.

    Args:
        adata: The inital AnnData object parsed by the :class: DataReader
        autodetect: Autodetection of categorical values (default: False)
        encodings: Only needed if autodetect set to False. A dict containing the categorical name
        and the encoding mode for the respective column

    Returns:
        An :class:`~anndata.AnnData` object with the encoded values in X
    """
    # check whether original layer exists; if not add it
    if "original" not in adata.layers.keys():
        adata.layers["original"] = adata.X.copy()
    # autodetect categorical values, which could lead to more categoricals
    if autodetect:
        if "var_to_encoding" in adata.uns.keys():
            print(
                "[bold yellow]The current AnnData object has been already encoded. Returning original AnnData object!"
            )
            return adata
        categoricals_names = _get_var_indices_for_type(adata, NON_NUMERIC_TAG)

        # no columns were detected, that would require an encoding (e.g. non numerical columns)
        if not categoricals_names:
            print("[bold yellow]Detected no columns that need to be encoded. Leaving passed AnnData object unchanged.")
            return adata
        # copy uns so it can be used in encoding process without mutating the original anndata object
        orig_uns_copy = adata.uns.copy()
        _add_categoricals_to_uns(adata, orig_uns_copy, categoricals_names)

        encoded_x = None
        encoded_var_names = adata.var_names.to_list()
        if encodings not in available_encodings - multi_encoding_modes:
            raise ValueError(
                f"Unknown encoding mode {encodings}. Please provide one of the following encoding modes:\n"
                f"{available_encodings - multi_encoding_modes}"
            )
        single_encode_mode_switcher = {
            "one-hot": _one_hot_encoding,
            "label": _label_encoding,
            "count": _count_encoding,
        }
        with Progress(
            "[progress.description]{task.description}",
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            refresh_per_second=1500,
        ) as progress:
            task = progress.add_task(f"[red]Running {encodings} on detected columns ...", total=1)
            # encode using the desired mode
            encoded_x, encoded_var_names = single_encode_mode_switcher[encodings](  # type: ignore
                adata,
                encoded_x,
                orig_uns_copy,
                encoded_var_names,
                categoricals_names,
                progress,
                task,
            )
            progress.update(task, description="Updating layer originals ...")
            # update layer content with the latest categorical encoding and the old other values
            logg.info("Encoding strings in X to save to .h5ad. Loading the file will reverse the encoding.")
            updated_layer = _update_layer_after_encoding(
                adata.layers["original"],
                encoded_x,
                encoded_var_names,
                adata.var_names.to_list(),
                categoricals_names,
            )
            progress.update(task, description=f"[bold blue]Finished {encodings} of autodetected columns.")

            # copy non-encoded columns, and add new tag for encoded columns. This is needed to track encodings
            new_var = pd.DataFrame(index=encoded_var_names)
            new_var[EHRAPY_TYPE_KEY] = adata.var[EHRAPY_TYPE_KEY].copy()
            new_var.loc[new_var.index.str.contains("ehrapycat")] = NON_NUMERIC_ENCODED_TAG

            encoded_ann_data = AnnData(
                encoded_x,
                obs=adata.obs.copy(),
                var=new_var,
                uns=orig_uns_copy,
                layers={"original": updated_layer},
            )
            encoded_ann_data.uns["var_to_encoding"] = {categorical: encodings for categorical in categoricals_names}
            encoded_ann_data.uns["encoding_to_var"] = {encodings: categoricals_names}

            _add_categoricals_to_obs(adata, encoded_ann_data, categoricals_names)

    # user passed categorical values with encoding mode for each of them
    else:
        # reencode data
        if "var_to_encoding" in adata.uns.keys():
            encodings = _reorder_encodings(adata, encodings)  # type: ignore
            adata = _undo_encoding(adata, "all")

        # are all specified encodings valid?
        for encoding in encodings.keys():  # type: ignore
            if encoding not in available_encodings:
                raise ValueError(
                    f"Unknown encoding mode {encoding}. Please provide one of the following encoding modes:\n"
                    f"{available_encodings}"
                )
        adata.uns["encoding_to_var"] = encodings

        categoricals_not_flat = list(chain(*encodings.values()))  # type: ignore
        # this is needed since multi column encoding will get passed a list of list instead of a flat list
        categoricals = list(
            chain(
                *(
                    _categoricals if isinstance(_categoricals, list) else (_categoricals,)
                    for _categoricals in categoricals_not_flat
                )
            )
        )
        # ensure no categorical column gets encoded twice
        if len(categoricals) != len(set(categoricals)):
            raise ValueError(
                "The categorical column names given contain at least one duplicate column. "
                "Check the column names to ensure that no column is encoded twice!"
            )
        elif any(cat in adata.var_names[adata.var[EHRAPY_TYPE_KEY] == NUMERIC_TAG] for cat in categoricals):
            print(
                "[bold yellow]At least one of passed column names seems to have numerical dtype. In general it is not recommended "
                "to encode numerical columns!"
            )
        orig_uns_copy = adata.uns.copy()
        _add_categoricals_to_uns(adata, orig_uns_copy, categoricals)
        var_to_encoding = {} if "var_to_encoding" not in adata.uns.keys() else adata.uns["var_to_encoding"]
        encoded_x = None
        encoded_var_names = adata.var_names.to_list()
        with Progress(
            "[progress.description]{task.description}",
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            refresh_per_second=1500,
        ) as progress:
            for encoding in encodings.keys():  # type: ignore
                task = progress.add_task(f"[red]Setting up {encodings}", total=1)
                encode_mode_switcher = {
                    "one-hot": _one_hot_encoding,
                    "label": _label_encoding,
                    "count": _count_encoding,
                    "hash": _hash_encoding,
                }
                progress.update(task, description=f"Running {encoding} ...")
                # perform the actual encoding
                encoded_x, encoded_var_names = encode_mode_switcher[encoding](
                    adata,
                    encoded_x,
                    orig_uns_copy,
                    encoded_var_names,
                    encodings[encoding],  # type: ignore
                    progress,
                    task,  # type: ignore
                )
                # update encoding history in uns
                for categorical in encodings[encoding]:  # type: ignore
                    # multi column encoding modes -> multiple encoded columns
                    if isinstance(categorical, list):
                        for column_name in categorical:
                            var_to_encoding[column_name] = encoding
                    else:
                        var_to_encoding[categorical] = encoding

        # update original layer content with the new categorical encoding and the old other values
        updated_layer = _update_layer_after_encoding(
            adata.layers["original"],
            encoded_x,
            encoded_var_names,
            adata.var_names.to_list(),
            categoricals,
        )

        # copy non-encoded columns, and add new tag for encoded columns. This is needed to track encodings
        new_var = pd.DataFrame(index=encoded_var_names)
        new_var[EHRAPY_TYPE_KEY] = adata.var[EHRAPY_TYPE_KEY].copy()
        new_var.loc[new_var.index.str.contains("ehrapycat")] = NON_NUMERIC_ENCODED_TAG

        try:
            encoded_ann_data = AnnData(
                X=encoded_x,
                obs=adata.obs.copy(),
                var=new_var,
                uns=orig_uns_copy,
                layers={"original": updated_layer},
            )
            # update current encodings in uns
            encoded_ann_data.uns["var_to_encoding"] = var_to_encoding

        # if the user did not pass every non numerical column for encoding, an Anndata object cannot be created
        except ValueError:
            raise AnnDataCreationError(
                "Creation of AnnData object failed. Ensure that you passed all non numerical, "
                "categorical values for encoding!"
            ) from None

        _add_categoricals_to_obs(adata, encoded_ann_data, categoricals)

    encoded_ann_data.X = encoded_ann_data.X.astype(np.number)

    logg.debug("Successfully encoded the AnnData object.")

    return encoded_ann_data


def _one_hot_encoding(
    adata: AnnData,
    X: np.ndarray | None,
    uns: dict[str, Any],
    var_names: list[str],
    categories: list[str],
    progress: Progress,
    task,
) -> tuple[np.ndarray, list[str]]:
    """Encode categorical columns using one hot encoding.

    Args:
        adata: The current AnnData object
        X: Current (encoded) X
        uns: A copy of the original uns
        var_names: Var names of current AnnData object
        categories: The name of the categorical columns to be encoded

    Returns:
        Encoded new X and the corresponding new var names
    """
    original_values = _initial_encoding(uns, categories)
    progress.update(task, description="[bold blue]Running one hot encoding on passed columns ...")
    encoder = OneHotEncoder(handle_unknown="ignore", sparse=False).fit(original_values)
    categorical_prefixes = [
        f"ehrapycat_{category}_{str(suffix).strip()}"
        for idx, category in enumerate(categories)
        for suffix in encoder.categories_[idx]
    ]
    transformed = encoder.transform(original_values)
    # X is None if this is the first encoding "round" -> take the former X
    if X is None:
        X = adata.X
    progress.advance(task, 1)
    progress.update(task, description="[blue]Updating X and var ...")

    temp_x, temp_var_names = _update_encoded_data(X, transformed, var_names, categorical_prefixes, categories)
    progress.update(task, description="[blue]Finished one hot encoding.")

    return temp_x, temp_var_names


def _label_encoding(
    adata: AnnData,
    X: np.ndarray | None,
    uns: dict[str, Any],
    var_names: list[str],
    categoricals: list[str],
    progress: Progress,
    task,
) -> tuple[np.ndarray, list[str]]:
    """Encode categorical columns using label encoding.

    Args:
        adata: The current AnnData object
        X: Current (encoded) X
        uns: A copy of the original uns
        var_names: Var names of current AnnData object
        categoricals: The name of the categorical columns, that need to be encoded

    Returns:
        Encoded new X and the corresponding new var names
    """
    original_values = _initial_encoding(uns, categoricals)
    # label encoding expects input array to be 1D, so iterate over all columns and encode them one by one
    for idx in range(original_values.shape[1]):
        progress.update(task, description=f"[blue]Running label encoding on column {categoricals[idx]} ...")
        label_encoder = LabelEncoder()
        row_vec = original_values[:, idx : idx + 1].ravel()  # type: ignore
        label_encoder.fit(row_vec)
        transformed = label_encoder.transform(row_vec)
        # need a column vector instead of row vector
        original_values[:, idx : idx + 1] = transformed[..., None]
        progress.advance(task, 1 / len(categoricals))
    category_prefixes = [f"ehrapycat_{categorical}" for categorical in categoricals]
    # X is None if this is the first encoding "round" -> take the former X
    if X is None:
        X = adata.X

    progress.update(task, description="[blue]Updating X and var ...")
    temp_x, temp_var_names = _update_encoded_data(X, original_values, var_names, category_prefixes, categoricals)
    progress.update(task, description="[blue]Finished label encoding.")

    return temp_x, temp_var_names


def _count_encoding(
    adata: AnnData,
    X: np.ndarray | None,
    uns: dict[str, Any],
    var_names: list[str],
    categoricals: list[str],
    progress: Progress,
    task,
) -> tuple[np.ndarray, list[str]]:
    """Encode categorical column using count encoding.

    Args:
        adata: The current AnnData object
        X: Current (encoded) X
        uns: A copy of the original uns
        var_names: Var names of current AnnData object
        categoricals: The name of the categorical columns, that need to be encoded

    Returns:
        Encoded new X and the corresponding new var names
    """
    original_values = _initial_encoding(uns, categoricals)
    progress.update(task, description="[blue]Running count encoding encoding on passed columns ...")
    # returns a pandas dataframe per default, but numpy array is needed
    count_encoder = CountEncoder(return_df=False)
    count_encoder.fit(original_values)
    category_prefix = [f"ehrapycat_{categorical}" for categorical in categoricals]
    transformed = count_encoder.transform(original_values)
    # X is None if this is the first encoding "round" -> take the former X
    if X is None:
        X = adata.X

    progress.advance(task, 1)
    progress.update(task, description="[blue]Updating X and var ...")
    temp_x, temp_var_names = _update_encoded_data(X, transformed, var_names, category_prefix, categoricals)
    progress.update(task, description="[blue]Finished count encoding.")

    return temp_x, temp_var_names


def _hash_encoding(
    adata: AnnData,
    X: np.ndarray | None,
    uns: dict[str, Any],
    var_names: list[str],
    categories: list[list[str]],
    progress: Progress,
    task,
) -> tuple[np.ndarray, list[str]]:
    """Encode categorical columns using hash encoding.

    Args:
        adata: The current AnnData object
        X: Current (encoded) X
        uns: A copy of the original uns
        var_names: Var names of current AnnData object
        categories: The name of the categorical columns to be encoded

    Returns:
        Encoded new X and the corresponding new var names
    """
    transformed_all, encoded_var_names = None, []
    for idx, multi_columns in enumerate(categories):
        progress.update(task, description=f"Running hash encoding on {idx + 1}. list ...")
        original_values = _initial_encoding(uns, multi_columns)

        encoder = HashingEncoder(return_df=False, n_components=8).fit(original_values)
        encoded_var_names += [f"ehrapycat_hash_{multi_columns[0]}" for _ in range(8)]
        transformed = encoder.transform(original_values)
        transformed_all = np.hstack((transformed_all, transformed)) if transformed_all is not None else transformed
        progress.advance(task, 1 / len(categories))

    # X is None if this is the first encoding "round" -> take the former X
    if X is None:
        X = adata.X
    progress.update(task, description="[blue]Updating X and var ...")

    temp_x, temp_var_names = _update_multi_encoded_data(
        X, transformed_all, var_names, encoded_var_names, sum(categories, [])
    )
    if temp_x.shape[1] != len(temp_var_names):
        raise HashEncodingError(
            "Hash encoding of input data failed. Note that hash encoding is not "
            "suitable for datasets with low number of data points and low cardinality!"
        )
    progress.update(task, description="[blue]Finished hash encoding.")

    return temp_x, temp_var_names


def _update_layer_after_encoding(
    old_layer: np.ndarray,
    new_x: np.ndarray,
    new_var_names: list[str],
    old_var_names: list[str],
    categories: list[str],
) -> np.ndarray:
    """Update the original layer containing the initial non categorical values and the latest encoded categoricals.

    Args:
        old_layer: The previous "original" layer
        new_x: The new encoded X
        new_var_names: The new encoded var names
        old_var_names: The previous var names
        categories: All previous categorical names

    Returns
        A Numpy array containing all numericals together with all encoded categoricals.
    """
    try:
        # get the index of the first column of the new encoded X, that does not store an encoded categorical
        new_cat_stop_index = next(i for i in range(len(new_var_names)) if not new_var_names[i].startswith("ehrapycat"))
        # get the index of the first column of the old encoded X, that does not store an encoded categorical
        old_cat_stop_index = next(i for i in range(len(old_var_names)) if not old_var_names[i].startswith("ehrapycat"))
    # when there are only encoded columns, simply return a copy of the new X, since to originals will be kept in the layer
    except StopIteration:
        return new_x.copy().astype("float32")
    # keep track of all indices with original value columns, that are (and were) not encoded
    idx_list = []
    for idx, col_name in enumerate(old_var_names[old_cat_stop_index:]):
        # this case is needed when there are one or more numerical (but categorical) columns that was not encoded yet
        if col_name not in categories:
            idx_list.append(idx + old_cat_stop_index)
    # slice old original layer using the selector
    old_layer_view = old_layer[:, idx_list]
    # get all encoded categoricals of X
    encoded_categoricals = new_x[:, :new_cat_stop_index]
    # horizontally stack all encoded categoricals and the remaining "old original values"
    updated_layer = np.hstack((encoded_categoricals, old_layer_view))

    try:
        logg.info("Updated the original layer after encoding.")
        return updated_layer.astype("float32")
    except ValueError as e:
        raise ValueError("Ensure that all columns which require encoding are being encoded.") from e


def _update_multi_encoded_data(
    X: np.ndarray,
    transformed: np.ndarray,
    var_names: list[str],
    encoded_var_names: list[str],
    categoricals: list[str],
) -> tuple[np.ndarray, list[str]]:
    """Update X and var_names after applying multi column encoding modes to some columns

    Args:
        X: Current (former) X
        transformed: The encoded (transformed) categorical columns
        var_names: Var names of current AnnData object
        encoded_var_names: The name(s) of the encoded column(s)
        categoricals: The categorical values that were encoded recently

    Returns:
        Encoded new X and the corresponding new var names
    """
    idx = []
    for pos, name in enumerate(var_names):
        if name in categoricals:
            idx.append(pos)
    # delete the original categorical column
    del_cat_column_x = np.delete(X, list(idx), 1)
    # create the new, encoded X
    temp_x = np.hstack((transformed, del_cat_column_x))
    # delete old categorical name
    var_names = [col_name for col_idx, col_name in enumerate(var_names) if col_idx not in idx]
    temp_var_names = encoded_var_names + var_names

    logg.debug("Updated `X` and `var_names` after applying multi column encoding modes.")

    return temp_x, temp_var_names


def _update_encoded_data(
    X: np.ndarray,
    transformed: np.ndarray,
    var_names: list[str],
    categorical_prefixes: list[str],
    categoricals: list[str],
) -> tuple[np.ndarray, list[str]]:
    """Update X and var_names after each encoding.

    Args:
        X: Current (former) X
        transformed: The encoded (transformed) categorical column
        var_names: Var names of current AnnData object
        categorical_prefixes: The name(s) of the encoded column(s)
        categoricals: The categorical values that were encoded recently

    Returns:
        Encoded new X and the corresponding new var names
    """
    idx = _get_categoricals_old_indices(var_names, categoricals)
    # delete the original categorical column
    del_cat_column_x = np.delete(X, list(idx), 1)
    # create the new, encoded X
    temp_x = np.hstack((transformed, del_cat_column_x))
    # delete old categorical name
    var_names = [col_name for col_idx, col_name in enumerate(var_names) if col_idx not in idx]
    temp_var_names = categorical_prefixes + var_names

    logg.debug("Updated `X` and `var_names` after each encoding.")

    return temp_x, temp_var_names


def _initial_encoding(
    uns: dict[str, Any],
    categoricals: list[str],
) -> np.ndarray:
    """Get all original values for all categoricals that need to be encoded (again).

    Args:
        uns: A copy of the original AnnData object's uns
        categoricals: All categoricals that need to be encoded

    Returns:
        Numpy array of all original categorial values
    """
    uns_: dict[str, np.ndarray] = uns
    # create numpy array from all original categorical values, that will be encoded (again)
    array = np.array(
        [uns_["original_values_categoricals"][categoricals[i]].ravel() for i in range(len(categoricals))]
    ).transpose()

    return array


def _undo_encoding(
    adata: AnnData,
    columns: str = "all",
    suppress_warning: bool = False,
) -> AnnData | None:
    """
    Undo the current encodings applied to all columns in X. This currently resets the AnnData object to its initial state.

    Args:
        adata: The AnnData object
        columns: The names of the columns to reset encoding for. Defaults to all columns. This resets the AnnData object to its initial state.
        suppress_warning: Whether warnings should be suppressed or not.

    Returns:
        A (partially) encoding reset AnnData object
    """
    if "var_to_encoding" not in adata.uns.keys():
        if not suppress_warning:
            print("[bold yellow]Calling undo_encoding on unencoded AnnData object.")
        return None

    # get all encoded variables
    encoded_categoricals = list(adata.uns["original_values_categoricals"].keys())
    # get all columns that should be stored in obs only
    columns_obs_only = [
        column_name for column_name in list(adata.obs.columns) if column_name not in encoded_categoricals
    ]

    if columns == "all":
        categoricals = list(adata.uns["original_values_categoricals"].keys())
    else:
        print("[bold yellow]Currently, one can only reset encodings for all columns! [bold red]Aborting...")
        return None
    transformed = _initial_encoding(adata.uns, categoricals)
    temp_x, temp_var_names = _delete_all_encodings(adata)
    new_x = np.hstack((transformed, temp_x)) if temp_x is not None else transformed
    new_var_names = categoricals + temp_var_names if temp_var_names is not None else categoricals
    # only keep columns in obs that were stored in obs only -> delete every encoded column from obs
    new_obs = adata.obs[columns_obs_only]
    uns = OrderedDict()
    # reset uns and keep numerical/non-numerical columns
    num_vars = _get_var_indices_for_type(adata, NUMERIC_TAG)
    non_num_vars = _get_var_indices_for_type(adata, NON_NUMERIC_TAG)
    for cat in categoricals:
        original_values = adata.uns["original_values_categoricals"][cat]
        type_first_nan = original_values[np.where(original_values != np.nan)][0]
        if isinstance(type_first_nan, (int, float, complex)) and not isinstance(type_first_nan, bool):
            num_vars.append(cat)
        else:
            non_num_vars.append(cat)

    var = pd.DataFrame(index=new_var_names)
    var[EHRAPY_TYPE_KEY] = NON_NUMERIC_TAG
    # Notice previously encoded columns are now newly added, and will stay tagged as non numeric
    var.loc[num_vars, EHRAPY_TYPE_KEY] = NUMERIC_TAG

    uns["numerical_columns"] = num_vars
    uns["non_numerical_columns"] = non_num_vars

    logg.debug("Encoding of the AnnData object was reset.")

    return AnnData(
        new_x,
        obs=new_obs,
        var=var,
        uns=uns,
        layers={"original": new_x.copy()},
    )


def _delete_all_encodings(adata: AnnData) -> tuple[np.ndarray | None, list | None]:
    """Delete all encoded columns and keep track of their indices.

    Args:
        adata: The AnnData object to operate on

    Returns:
        A temporary X were all encoded columns are deleted and all var_names of unencoded columns.
    """
    var_names = list(adata.var_names)
    if adata.X is not None and var_names is not None:
        idx = 0
        for var in var_names:
            if not var.startswith("ehrapycat"):
                break
            idx += 1
        # case: only encoded columns were found
        if idx == len(var_names):
            return None, None
        # don't need to consider case when no encoded columns are there, since undo_encoding would not run anyways in this case
        logg.info("All encoded columns of the AnnData object were deleted.")
        return adata.X[:, idx:].copy(), var_names[idx:]
    return None, None


def _reorder_encodings(adata: AnnData, new_encodings: dict[str, list[list[str]] | list[str]]):
    """Reorder the encodings and update which column will be encoded using which mode (with which columns in case of multi column encoding modes).

    Args:
        adata: The AnnData object to be reencoded
        new_encodings: The new encodings passed by the user (might affect encoded as well as previously non encoded columns)

    Returns:
        An updated encoding scheme
    """
    flattened_modes: list[list[str] | str] = sum(new_encodings.values(), [])  # type: ignore
    latest_encoded_columns = list(chain(*(i if isinstance(i, list) else (i,) for i in flattened_modes)))
    # check for duplicates and raise an error if any
    if len(set(latest_encoded_columns)) != len(latest_encoded_columns):
        print(
            "[bold red]Reencoding AnnData object failed. You have at least one duplicate in your encodings. A column "
            "cannot be encoded at the same time using different encoding modes!"
        )
        raise DuplicateColumnEncodingError
    old_encode_mode = adata.uns["var_to_encoding"]
    for categorical in latest_encoded_columns:
        encode_mode = old_encode_mode.get(categorical)
        # if None, this categorical has not been encoded before but will be encoded now for the first time
        # multi column encoding mode
        if encode_mode in multi_encoding_modes:
            encoded_categoricals_with_mode = adata.uns["encoding_to_var"][encode_mode]
            _found = False
            for column_list in encoded_categoricals_with_mode:
                for column_name in column_list:
                    if column_name == categorical:
                        column_list.remove(column_name)
                        _found = True
                        break
                # a categorical can only be encoded once and therefore found once
                if _found:
                    break
            # filter all lists that are now empty since all variables will be reencoded from this list
            updated_multi_list = filter(None, encoded_categoricals_with_mode)
            # if no columns remain that will be encoded with this encode mode, delete this mode from modes as well
            if not list(updated_multi_list):
                del adata.uns["encoding_to_var"][encode_mode]
        # single column encoding mode
        elif encode_mode in available_encodings:
            encoded_categoricals_with_mode = adata.uns["encoding_to_var"][encode_mode]
            for ind, column_name in enumerate(encoded_categoricals_with_mode):
                if column_name == categorical:
                    del encoded_categoricals_with_mode[ind]
                    break
                # if encoding mode is
            if not encoded_categoricals_with_mode:
                del adata.uns["encoding_to_var"][encode_mode]

    return _update_new_encode_modes(new_encodings, adata.uns["encoding_to_var"])


def _update_new_encode_modes(
    new_encodings: dict[str, list[list[str]] | list[str]],
    filtered_old_encodings: dict[str, list[list[str]] | list[str]],
):
    """Update the encoding scheme.

    If the encoding mode exists in the filtered old encodings, append all values (columns that should be encoded using this mode) to this key.
    If not, defaultdict ensures that no KeyError will be raised and the values are simply appended to the default value ([]).

    Args:
        new_encodings: The new encoding modes passed by the user; basically what will be passed for encodings when calling the encode API
        filtered_old_encodings: The old encoding modes, but with all columns removed that will be reencoded

    Returns:
        The updated encoding scheme
    """
    updated_encodings = defaultdict(list)  # type: ignore
    for k, v in chain(new_encodings.items(), filtered_old_encodings.items()):
        updated_encodings[k] += v

    logg.debug("Updated the encoding scheme.")

    return dict(updated_encodings)


def _get_categoricals_old_indices(old_var_names: list[str], encoded_categories: list[str]) -> set[int]:
    """Get indices of every (possibly encoded) categorical column belonging to a newly encoded categorical value.

    Args:
        old_var_names: Former variables names
        encoded_categories: Already encoded categories

    Returns:
        Set of all indices of formerly encoded categories belonging to a newly encoded categorical value.
    """
    idx_list = set()
    category_set = set(encoded_categories)
    for idx, old_var_name in enumerate(old_var_names):
        # if the old variable was previously unencoded (only the case for numerical categoricals)
        if old_var_name in category_set:
            idx_list.add(idx)
        # if the old variable was already encoded
        elif old_var_name.startswith("ehrapycat_"):
            if any(old_var_name[10:].startswith(category) for category in category_set):
                idx_list.add(idx)

    return idx_list


def _add_categoricals_to_obs(original: AnnData, new: AnnData, categorical_names: list[str]) -> None:
    """Add the original categorical values to obs.

    Args:
        original: The original AnnData object
        new: The new AnnData object
        categorical_names: Name of each categorical column
    """
    for idx, var_name in enumerate(original.var_names):
        if var_name in new.obs.columns:
            continue
        elif var_name in categorical_names:
            new.obs[var_name] = original.X[::, idx : idx + 1].flatten()
            # note: this will count binary columns (0 and 1 only) as well
            # needed for writing to .h5ad files
            if set(pd.unique(new.obs[var_name])).issubset({False, True, np.NaN}):
                new.obs[var_name] = new.obs[var_name].astype("bool")
    # get all non bool object columns and cast the to category dtype
    object_columns = list(new.obs.select_dtypes(include="object").columns)
    new.obs[object_columns] = new.obs[object_columns].astype("category")
    logg.info(f"The original categorical values `{categorical_names}` were added to obs.")


def _add_categoricals_to_uns(original: AnnData, new: AnnData, categorical_names: list[str]) -> None:
    """Add the original categorical values to uns.

    Args:
        original: The original AnnData object
        new: The new AnnData object
        categorical_names: Name of each categorical column
    """
    is_initial = "original_values_categoricals" in original.uns.keys()
    new["original_values_categoricals"] = {} if not is_initial else original.uns["original_values_categoricals"].copy()

    for idx, var_name in enumerate(original.var_names):
        if is_initial and var_name in new["original_values_categoricals"]:
            continue
        elif var_name in categorical_names:
            # keep numerical dtype when writing original values to uns
            if var_name in original.var_names[original.var[EHRAPY_TYPE_KEY] == NUMERIC_TAG]:
                new["original_values_categoricals"][var_name] = original.X[::, idx : idx + 1].astype("float")
            else:
                new["original_values_categoricals"][var_name] = original.X[::, idx : idx + 1].astype("str")

    logg.info(f"The original categorical values `{categorical_names}` were added to uns.")


class AlreadyEncodedWarning(UserWarning):
    pass


class AnnDataCreationError(ValueError):
    pass


class DuplicateColumnEncodingError(ValueError):
    pass


class HashEncodingError(Exception):
    pass



import ehrapy as ep
from ehrapy.preprocessing._highly_variable_features import highly_variable_features


def test_highly_variable_features():
    adata = ep.dt.dermatology(encoded=True)
    ep.pp.knn_impute(adata)
    highly_variable_features(adata)

    assert "highly_variable" in adata.var.columns
    assert "highly_variable_rank" in adata.var.columns
    assert "means" in adata.var.columns
    assert "variances" in adata.var.columns
    assert "variances_norm" in adata.var.columns

    adata = ep.dt.dermatology(encoded=True)
    ep.pp.knn_impute(adata)
    highly_variable_features(adata, top_features_percentage=0.5)
    assert adata.var["highly_variable"].sum() == 17



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal

import fhiry.parallel as fp
import numpy as np
import pandas as pd
from _collections import OrderedDict
from anndata import AnnData
from anndata import read as read_h5
from rich import print

from ehrapy import ehrapy_settings, settings
from ehrapy.anndata.anndata_ext import df_to_anndata
from ehrapy.data._dataloader import download, remove_archive_extension
from ehrapy.preprocessing._encode import encode

if TYPE_CHECKING:
    from collections.abc import Iterator


def read_csv(
    dataset_path: Path | str,
    sep: str = ",",
    index_column: dict[str, str | int] | str | int | None = None,
    columns_obs_only: dict[str, list[str]] | list[str] | None = None,
    columns_x_only: dict[str, list[str]] | list[str] | None = None,
    return_dfs: bool = False,
    cache: bool = False,
    backup_url: str | None = None,
    download_dataset_name: str | None = None,
    archive_format: Literal["zip", "tar", "tar.gz", "tgz"] = None,
    **kwargs,
) -> AnnData | dict[str, AnnData]:
    """Reads or downloads a desired directory of csv/tsv files or a single csv/tsv file.

    Args:
        dataset_path: Path to the file or directory to read.
        sep: Separator in the file. One of either , (comma) or \t (tab). Defaults to , (comma)
        index_column: The index column of obs. Usually the patient visit ID or the patient ID.
        columns_obs_only: These columns will be added to obs only and not X.
        columns_x_only: These columns will be added to X only and all remaining columns to obs.
                        Note that datetime columns will always be added to .obs though.
        return_dfs: Whether to return one or several Pandas DataFrames.
        cache: Whether to write to cache when reading or not. Defaults to False .
        download_dataset_name: Name of the file or directory after download.
        backup_url: URL to download the data file(s) from, if the dataset is not yet on disk.
        is_archive: Whether the downloaded file is an archive.

    Returns:
        An :class:`~anndata.AnnData` object or a dict with an identifier (the filename, without extension)
        for each :class:`~anndata.AnnData` object in the dict

    Examples:
        >>> import ehrapy as ep
        >>> adata = ep.io.read_csv("myfile.csv")
    """
    _check_columns_only_params(columns_obs_only, columns_x_only)
    dataset_path = Path(dataset_path)
    if not dataset_path.exists():
        dataset_path = _get_non_existing_files(dataset_path, download_dataset_name, backup_url, archive_format)

    adata = _read_csv(
        file_path=dataset_path,
        sep=sep,
        index_column=index_column,
        columns_obs_only=columns_obs_only,
        columns_x_only=columns_x_only,
        return_dfs=return_dfs,
        cache=cache,
        **kwargs,
    )
    return adata


def _read_csv(
    file_path: Path,
    sep: str,
    index_column: dict[str, str | int] | str | int | None,
    columns_obs_only: dict[str, list[str]] | list[str] | None,
    columns_x_only: dict[str, list[str]] | list[str] | None,
    return_dfs: bool = False,
    cache: bool = False,
    **kwargs,
) -> AnnData | dict[str, AnnData]:
    """Internal interface of the read_csv method."""
    if cache and return_dfs:
        raise CachingNotSupported("Caching is currently not supported for Pandas DataFrame objects.")
    if return_dfs and (columns_x_only or columns_obs_only):
        raise Warning(
            "Parameters columns_x_only and columns_obs_only are not supported when returning Pandas DataFrames."
        )

    path_cache = settings.cachedir / file_path
    # reading from (cache) file is separated in the read_h5ad function
    if cache and (path_cache.is_dir() or path_cache.is_file()):
        raise CacheExistsException(
            f"{path_cache} already exists. Use the read_h5ad function instead to read from cache!"
        )

    # If the the file path is a directory, assume it is a dataset with multiple files
    elif file_path.is_dir():
        return _read_from_directory(
            file_path,
            cache,
            path_cache,
            extension=sep,
            index_column=index_column,
            columns_obs_only=columns_obs_only,
            columns_x_only=columns_x_only,
            return_dfs=return_dfs,
        )
    # input is a single file
    else:
        if sep not in {",", "\t"}:
            raise ValueError("Please provide one of the available separators , or tab")
        adata, columns_obs_only = _do_read_csv(
            file_path,
            sep,
            index_column,  # type: ignore
            columns_obs_only,  # type: ignore
            columns_x_only,  # type: ignore
            cache,
            **kwargs,
        )
        # cache results if desired
        if cache:
            if not path_cache.parent.is_dir():
                path_cache.parent.mkdir(parents=True)
            return _write_cache(adata, path_cache, columns_obs_only)  # type: ignore
        return adata


def read_h5ad(
    dataset_path: Path | str,
    backup_url: str | None = None,
    download_dataset_name: str | None = None,
    archive_format: Literal["zip", "tar", "tar.gz", "tgz"] = None,
) -> AnnData | dict[str, AnnData]:
    """Reads or downloads a desired directory of h5ad files or a single h5ad file.

    Args:
        dataset_path: Path to the file or directory to read.
        download_dataset_name: Name of the file or directory in case the dataset is downloaded
        backup_url: URL to download the data file(s) from if not yet existing.

    Returns:
        An :class:`~anndata.AnnData` object or a dict with an identifier (the filename, without extension)
        for each :class:`~anndata.AnnData` object in the dict

    Examples:
        >>> import ehrapy as ep
        >>> adata = ep.dt.mimic_2(encoded=True)
        >>> ep.io.write("mimic_2.h5ad", adata)
        >>> adata_2 = ep.io.read_h5ad("mimic_2.h5ad")
    """
    file_path: Path = Path(dataset_path)
    if not file_path.exists():
        file_path = _get_non_existing_files(file_path, download_dataset_name, backup_url, archive_format=archive_format)

    if file_path.is_dir():
        adata = _read_from_directory(file_path, False, None, "h5ad")
    else:
        adata = _do_read_h5ad(file_path)

    return adata


def _read_from_directory(
    file_path: Path,
    cache: bool,
    path_cache_dir: Path | None,
    extension: str,
    index_column: dict[str, str | int] | str | int | None = None,
    columns_obs_only: dict[str, list[str]] | list[str] | None = None,
    columns_x_only: dict[str, list[str]] | list[str] | None = None,
    return_dfs: bool = False,
) -> dict[str, AnnData] | dict[str, pd.DataFrame]:
    """Parse AnnData objects or Pandas DataFrames from a directory containing the data files"""
    if return_dfs:
        dfs = _read_multiple_csv(file_path, sep=extension, return_dfs=True)
        return dfs  # type: ignore
    if extension in {",", "\t"}:
        adata_objects, columns_obs_only = _read_multiple_csv(  # type: ignore
            file_path,
            sep=extension,
            index_column=index_column,
            columns_obs_only=columns_obs_only,
            columns_x_only=columns_x_only,
            return_dfs=False,
        )
        # cache results
        if cache:
            if not path_cache_dir.parent.is_dir():
                path_cache_dir.parent.mkdir(parents=True)
            path_cache_dir.mkdir()
            return _write_cache_dir(adata_objects, path_cache_dir, columns_obs_only, index_column)  # type: ignore
        return adata_objects  # type: ignore
    elif extension == "h5ad":
        return _read_multiple_h5ad(file_path)
    else:
        raise NotImplementedError(f"Reading from directory with .{extension} files is not implemented yet!")


def _read_multiple_csv(
    file_path: Path,
    sep: str,
    index_column: dict[str, str | int] | str | int | None = None,
    columns_obs_only: dict[str, list[str]] | list[str] | None = None,
    columns_x_only: dict[str, list[str]] | list[str] | None = None,
    return_dfs: bool = False,
    cache: bool = False,
    **kwargs,
) -> tuple[dict[str, AnnData], dict[str, list[str] | None]] | dict[str, pd.DataFrame]:
    """Read a dataset containing multiple .csv/.tsv files.

    Args:
        file_path: File path to the directory containing multiple .csv/.tsv files.
        sep: Either , or \t to determine which files to read.
        index_column: Column names of the index columns for obs
        columns_obs_only: List of columns per file (AnnData object) which should only be stored in .obs, but not in X.
                          Useful for free text annotations.
        columns_x_only: List of columns per file (AnnData object) which should only be stored in .X, but not in obs.
                        Datetime columns will be added to .obs regardless.
        return_dfs: When set to True, return a dictionary of Pandas DataFrames.
        cache: Whether to cache results or not
        kwargs: Keyword arguments for Pandas `read_csv`

    Returns:
        A Dict mapping the filename (object name) to the corresponding :class:`~anndata.AnnData` object and the columns
        that are obs only for each object
    """
    obs_only_all = {}
    if return_dfs:
        df_dict: dict[str, pd.DataFrame] = {}
    else:
        anndata_dict = {}

    for file in file_path.iterdir():
        if file.is_file() and file.suffix in {".csv", ".tsv"}:
            # slice off the file suffix .csv or .tsv for a clean file name
            file_identifier = file.name[:-4]
            if return_dfs:
                df = pd.read_csv(file, sep=sep, **kwargs)
                df_dict[file_identifier] = df
                continue

            index_col, col_obs_only, col_x_only = _extract_index_and_columns_obs_only(
                file_identifier, index_column, columns_obs_only, columns_x_only
            )
            adata, single_adata_obs_only = _do_read_csv(file, sep, index_col, col_obs_only, col_x_only, cache=cache)
            obs_only_all[file_identifier] = single_adata_obs_only
            # obs indices have to be unique otherwise updating and working with the object will fail
            if index_col:
                adata.obs_names_make_unique()

            anndata_dict[file_identifier] = adata
    if return_dfs:
        return df_dict
    else:
        return anndata_dict, obs_only_all


def _do_read_csv(
    file_path: Path | Iterator[str],
    delimiter: str | None = ",",
    index_column: str | int | None = None,
    columns_obs_only: list[str] | None = None,
    columns_x_only: list[str] | None = None,
    cache: bool = False,
    **kwargs,
) -> tuple[AnnData, list[str] | None]:
    """Read `.csv` and `.tsv` file.

    Args:
        file_path: File path to the csv file.
        delimiter: Delimiter separating the csv data within the file.
        index_column: Index or column name of the index column (obs)
        columns_obs_only: List of columns which only be stored in .obs, but not in X. Useful for free text annotations.
        columns_x_only: List of columns which only be stored in X, but not in .obs.

        cache: Whether the data should be written to cache or not

    Returns:
        An :class:`~anndata.AnnData` object and the column obs only for the object
    """
    try:
        if index_column and columns_obs_only and index_column in columns_obs_only:
            print(
                f"[bold yellow]Index column [blue]{index_column} [yellow]is also used as a column "
                f"for obs only. Using default indices instead and moving [blue]{index_column} [yellow]to column_obs_only."
            )
            index_column = None
        initial_df = pd.read_csv(file_path, delimiter=delimiter, index_col=index_column, **kwargs)
    # in case the index column is misspelled or does not exist
    except ValueError:
        raise IndexNotFoundError(
            f"Could not create AnnData object while reading file {file_path} . Does index_column named {index_column} "
            f"exist in {file_path}?"
        ) from None

    initial_df, columns_obs_only = _prepare_dataframe(initial_df, columns_obs_only, columns_x_only, cache)

    return df_to_anndata(initial_df, columns_obs_only), columns_obs_only


def _read_multiple_h5ad(
    file_path: Path,
) -> dict[str, AnnData]:
    """Read a dataset containing multiple .h5ad files.

    Args:
        file_path: File path to the directory containing multiple .csv/.tsv files.

    Returns:
        A dict mapping the filename (object name) to the corresponding :class:`~anndata.AnnData` object
    """
    anndata_dict = {}
    for file in file_path.iterdir():
        if file.is_file() and file.suffix == ".h5ad":
            # slice off the file suffix .h5ad
            adata_identifier = file.name[:-5]
            adata = _do_read_h5ad(file)
            anndata_dict[adata_identifier] = adata
    return anndata_dict


def _do_read_h5ad(file_path: Path | Iterator[str]) -> AnnData:
    """Read from a h5ad file.
    Args:
        file_path: Path to the h5ad file

    Returns:
        An AnnData object.
    """
    import anndata as ad

    adata = ad.read_h5ad(file_path)
    if "ehrapy_dummy_encoding" in adata.uns.keys():
        # if dummy encoding was needed, the original dtype of X could not be numerical, so cast it to object
        adata.X = adata.X.astype("object")
        decoded_adata = _decode_cached_adata(adata, list(adata.uns["columns_obs_only"]))
        return decoded_adata
    return adata


def read_fhir(
    dataset_path: str,
    format: Literal["json", "ndjson"] = "json",
    columns_obs_only: list[str] | None = None,
    columns_x_only: list[str] | None = None,
    return_df: bool = False,
    cache: bool = False,
    backup_url: str | None = None,
    index_column: str | int | None = None,
    download_dataset_name: str | None = None,
    archive_format: Literal["zip", "tar", "tar.gz", "tgz"] = None,
) -> pd.DataFrame | AnnData:
    """Reads one or multiple FHIR files using fhiry.

    Uses https://github.com/dermatologist/fhiry to read the FHIR file into a Pandas DataFrame
    which is subsequently transformed into an AnnData object.

    Be aware that FHIR data can be nested and return lists or dictionaries as values.
    In such cases, one can either:
    1. Transform the data into an awkward array and flatten it when needed.
    2. Extract values from all lists and dictionaries to store single values in the fields.
    3. Remove all lists and dictionaries. Only do this if the information is not relevant to you.

    Args:
        dataset_path: Path to one or multiple FHIR files.
        format: The file format of the FHIR data. One of 'json' or 'ndjson'. Defaults to 'json'.
        columns_obs_only: These columns will be added to obs only and not X.
        columns_x_only: These columns will be added to X only and all remaining columns to obs.
                        Note that datetime columns will always be added to .obs though.
        return_df: Whether to return one or several Pandas DataFrames.
        cache: Whether to write to cache when reading or not. Defaults to False.
        download_dataset_name: Name of the file or directory in case the dataset is downloaded
        index_column: The index column for the generated object. Usually the patient or visit ID.
        backup_url: URL to download the data file(s) from if not yet existing.

    Returns:
        A Pandas DataFrame or AnnData object of the read in FHIR file(s).

    Examples:
        >>> import ehrapy as ep
        >>> adata = ep.io.read_fhir("/path/to/fhir/resources")

        Be aware that most FHIR datasets have nested data that might need to be removed.
        In such cases consider working with DataFrames.
        >>> df = ep.io.read_fhir("/path/to/fhir/resources", return_df=True)
        >>> df.drop(columns=[col for col in df.columns if any(isinstance(x, (list, dict)) for x in df[col].dropna())], inplace=True)
        >>> df.drop(columns=df.columns[df.isna().all()], inplace=True)
    """
    _check_columns_only_params(columns_obs_only, columns_x_only)
    file_path: Path = Path(dataset_path)
    if not file_path.exists():
        file_path = _get_non_existing_files(file_path, download_dataset_name, backup_url, archive_format)

    adata = _read_fhir(
        file_path=str(file_path.resolve()),
        format=format,
        index_column=index_column,
        columns_obs_only=columns_obs_only,
        columns_x_only=columns_x_only,
        return_df=return_df,
        cache=cache,
    )
    return adata


def _read_fhir(
    file_path: str,
    format: Literal["json", "ndjson"],
    index_column: dict[str, str | int] | str | int | None,
    columns_obs_only: list[str] | None,
    columns_x_only: list[str] | None,
    return_df: bool = False,
    cache: bool = False,
) -> AnnData | dict[str, AnnData]:
    """Internal interface of the read_fhir method."""
    if cache and return_df:
        raise CachingNotSupported("Caching is currently not supported for or Pandas DataFrame objects.")
    if return_df and (columns_x_only or columns_obs_only):
        raise Warning(
            "Parameters columns_x_only and columns_obs_only are not supported when returning Pandas DataFrames."
        )
    path_cache = settings.cachedir / file_path
    if cache and (path_cache.is_dir() or path_cache.is_file()):
        raise CacheExistsException(
            f"{path_cache} already exists. Use the read_h5ad function instead to read from cache!"
        )
    if format == "json":
        df = fp.process(file_path)
    elif format == "ndjson":
        df = fp.ndjson(file_path)
    else:
        raise ValueError("Only folders containing json and ndjson in FHIR format are supported.")

    df, columns_obs_only = _prepare_dataframe(df, columns_obs_only, columns_x_only, cache)
    if index_column:
        df.set_index(index_column)

    if return_df:
        return df
    else:
        adata = df_to_anndata(df, columns_obs_only)

    if cache:
        if not path_cache.parent.is_dir():
            path_cache.parent.mkdir(parents=True)
        return _write_cache(adata, path_cache, columns_obs_only)  # type: ignore

    return adata


def _get_non_existing_files(
    dataset_path: Path,
    download_dataset_name: str,
    backup_url: str,
    archive_format: Literal["zip", "tar", "tar.gz", "tgz"] = None,
) -> Path:
    """Handle non existing files or directories by trying to download from a backup_url and moving them in the correct directory.

    Returns:
        The file or directory path of the downloaded content.
    """
    if backup_url is None and not dataset_path.exists():
        raise ValueError(
            f"File or directory {dataset_path} does not exist and no backup_url was provided.\n"
            f"Please provide a backup_url or check whether path is spelled correctly."
        )
    print("[bold yellow]Path or dataset does not yet exist. Attempting to download...")
    download(
        backup_url,
        output_file_name=download_dataset_name,
        output_path=ehrapy_settings.datasetdir,
        archive_format=archive_format,
    )

    if archive_format:
        dataset_path = remove_archive_extension(dataset_path)

    return dataset_path


def _read_from_cache_dir(cache_dir: Path) -> dict[str, AnnData]:
    """Read AnnData objects from the cache directory."""
    adata_objects = {}
    # read each cache file in the cache directory and store it into a dict
    for cache_file in cache_dir.iterdir():
        if cache_file.name.endswith(".h5ad"):
            adata_objects[cache_file.stem] = _read_from_cache(cache_file)
    return adata_objects


def _read_from_cache(path_cache: Path) -> AnnData:
    """Read AnnData object from cached file."""
    cached_adata = read_h5(path_cache)
    # type cast required when dealing with non numerical data; otherwise all values in X would be treated as strings
    if not np.issubdtype(cached_adata.X.dtype, np.number):
        cached_adata.X = cached_adata.X.astype("object")
    try:
        columns_obs_only = list(cached_adata.uns["columns_obs_only"])
        del cached_adata.uns["columns_obs_only"]
    # in case columns_obs_only has not been passed
    except KeyError:
        columns_obs_only = []
    # required since reading from cache returns a numpy array instead of a list here
    cached_adata.uns["numerical_columns"] = list(cached_adata.uns["numerical_columns"])
    # recreate the original AnnData object with the index column for obs and obs only columns
    cached_adata = _decode_cached_adata(cached_adata, columns_obs_only)

    return cached_adata


def _write_cache_dir(
    adata_objects: dict[str, AnnData],
    path_cache: Path,
    columns_obs_only,
    index_column: dict[str, str | int] | None,  # type ignore
) -> dict[str, AnnData]:
    """Write multiple AnnData objects into a common cache directory keeping index column and columns_obs_only.

    Args:
        adata_objects: A dictionary with an identifier as key for each of the AnnData objects
        path_cache: Path to the cache directory
        columns_obs_only: Columns for obs only
        index_column: The index columns for each object (if any)

    Returns:
        A dict containing an unique identifier and an :class:`~anndata.AnnData` object for each file read
    """
    for identifier in adata_objects:
        # for each identifier (for the AnnData object), we need the index column and obs_only cols (if any) for reuse when reading cache
        index_col, cols_obs_only, _ = _extract_index_and_columns_obs_only(identifier, index_column, columns_obs_only)
        adata_objects[identifier] = _write_cache(
            adata_objects[identifier], path_cache / (identifier + ".h5ad"), cols_obs_only
        )
    return adata_objects


def _write_cache(
    raw_anndata: AnnData,
    path_cache: Path,
    columns_obs_only: list[str] | None,
) -> AnnData:
    """Write AnnData object to cache"""
    original_x_dtype = raw_anndata.X.dtype
    if not np.issubdtype(original_x_dtype, np.number):
        cached_adata = encode(data=raw_anndata, autodetect=True)
    else:
        cached_adata = raw_anndata
    # temporary key that stores all column names that are obs only for this AnnData object
    cached_adata.uns["columns_obs_only"] = columns_obs_only
    cached_adata.uns["ehrapy_dummy_encoding"] = True
    # append correct file suffix
    if not path_cache.suffix == ".h5ad":
        if path_cache.suffix in {".tsv", ".csv"}:
            path_cache = Path(str(path_cache)[:-4] + ".h5ad")
        else:
            path_cache = Path(str(path_cache) + ".h5ad")
    cached_adata.write(path_cache)
    # preserve original dtype of X (either numerical or object)
    cached_adata.X = cached_adata.X.astype(original_x_dtype)
    cached_adata = _decode_cached_adata(cached_adata, columns_obs_only)
    return cached_adata


def _prepare_dataframe(initial_df: pd.DataFrame, columns_obs_only, columns_x_only=None, cache=False):
    """Prepares the dataframe to be casted into an AnnData object.

    Datetime columns will be detected and added to columns_obs_only.

    Returns:
         The initially parsed dataframe and an updated list of columns_obs_only.
    """
    # when passing columns x only, simply handle the (asymmetric) difference to be obs only and everything else is kept in X
    if columns_x_only:
        columns_obs_only = list(set(initial_df.columns) - set(columns_x_only))
    # get all object dtype columns
    object_type_columns = [col_name for col_name in initial_df.columns if initial_df[col_name].dtype == "object"]
    # if columns_obs_only is None, initialize it as datetime columns need to be included here
    if not columns_obs_only:
        columns_obs_only = []
    no_datetime_object_col = []
    for col in object_type_columns:
        try:
            pd.to_datetime(initial_df[col], format="mixed")
            # only add to column_obs_only if not present already to avoid duplicates
            if col not in columns_obs_only:
                columns_obs_only.append(col)
        except (ValueError, TypeError):
            # we only need to replace NANs on non datetime, non numerical columns since datetime are obs only by default
            no_datetime_object_col.append(col)
    # writing to hd5a files requires non string to be empty in non numerical columns
    if cache:
        # TODO remove this when anndata 0.8.0 is released
        initial_df[no_datetime_object_col] = initial_df[no_datetime_object_col].fillna("")
        # temporary workaround needed; see https://github.com/theislab/anndata/issues/504 and https://github.com/theislab/anndata/issues/662
        # converting booleans to strings is needed for caching as writing to .h5ad files currently does not support writing boolean values
        bool_columns = {
            column_name: "str" for column_name in initial_df.columns if initial_df.dtypes[column_name] == "bool"
        }
        initial_df = initial_df.astype(bool_columns)
    return initial_df, columns_obs_only


def _decode_cached_adata(adata: AnnData, column_obs_only: list[str]) -> AnnData:
    """Decode the label encoding of initial AnnData object

    Args:
        adata: The label encoded AnnData object
        column_obs_only: The columns, that should be kept in obs

    Returns:
        The decoded, initial AnnData object
    """
    var_names = list(adata.var_names)
    # for each encoded categorical, replace its encoded values with its original values in X
    for idx, var_name in enumerate(var_names):
        if not var_name.startswith("ehrapycat_"):
            break
        value_name = var_name[10:]
        original_values = adata.uns["original_values_categoricals"][value_name]
        adata.X[:, idx : idx + 1] = original_values
        # update var name per categorical
        var_names[idx] = value_name
    # drop all columns, that are not obs only in obs
    if column_obs_only:
        adata.obs = adata.obs[column_obs_only]
    else:
        adata.obs = pd.DataFrame(index=adata.obs.index)
    # set the new var names (unencoded ones)
    adata.var.index = var_names
    adata.layers["original"] = adata.X.copy()
    # reset uns but keep numerical columns
    numerical_columns = adata.uns["numerical_columns"]
    adata.uns = OrderedDict()
    adata.uns["numerical_columns"] = numerical_columns
    adata.uns["non_numerical_columns"] = list(set(adata.var_names) ^ set(numerical_columns))

    return adata


def _extract_index_and_columns_obs_only(identifier: str, index_columns, columns_obs_only, columns_x_only=None):
    """Extract the index column (if any) and the columns, for obs only (if any) from the given user input.

    For each file, `index_columns` and `columns_obs_only` can provide three cases:
        1.) The filename (thus the identifier) is not present as a key and no default key is provided or one or both dicts are empty:
            --> No index column will be set and/or no columns are obs only (based on user input)

        2.) The filename (thus the identifier) is not present as a key, but default key is provided
            --> The index column will be set and/or columns will be obs only according to the default key

        3.) The filename is present as a key
            --> The index column will be set and/or columns are obs only according to its value

    Args:
        identifier: The name of the
        index_columns: Index columns
        columns_obs_only: Columns for obs only

    Returns:
        Index column (if any) and columns obs only (if any) for this specific AnnData object
    """
    _index_column = None
    _columns_obs_only = None
    _columns_x_only = None
    # get index column (if any)
    if index_columns and identifier in index_columns.keys():
        _index_column = index_columns[identifier]
    elif index_columns and "default" in index_columns.keys():
        _index_column = index_columns["default"]

    # get columns obs only (if any)
    if columns_obs_only and identifier in columns_obs_only.keys():
        _columns_obs_only = columns_obs_only[identifier]
    elif columns_obs_only and "default" in columns_obs_only.keys():
        _columns_obs_only = columns_obs_only["default"]

    # get columns x only (if any)
    if columns_x_only and identifier in columns_x_only.keys():
        _columns_x_only = columns_x_only[identifier]
    elif columns_x_only and "default" in columns_x_only.keys():
        _columns_x_only = columns_x_only["default"]

    # if index column is also found in column_obs_only or x_only, use default indices instead and only move it to obs/X, but warn the user
    if (_index_column and _columns_obs_only or _index_column and _columns_x_only) and (
        _index_column in _columns_obs_only or _index_column in _columns_x_only
    ):
        print(
            f"[bold yellow]Index column [blue]{_index_column} [yellow]for file [blue]{identifier} [yellow]is also used as a column "
            f"for obs or X only. Using default indices instead and moving [blue]{_index_column} [yellow]to obs/X!."
        )
        _index_column = None

    return _index_column, _columns_obs_only, _columns_x_only


def _check_columns_only_params(
    obs_only: dict[str, list[str]] | list[str] | None, x_only: dict[str, list[str]] | list[str] | None
) -> None:
    """Check whether columns_obs_only and columns_x_only are passed exclusively.

    For a single AnnData object (thus parameters being a list of strings) it's not desirable to pass both, obs_only and x_only.
    For multiple AnnData objects (thus the parameters being dicts of string keys with a list value), it is possible to pass both. But the keys
    (unique identifiers of the AnData objects, basically its names) should share no common identifier,
    thus a single AnnData object is either in x_only OR obs_only, but not in both.
    """
    if not obs_only or not x_only:
        return
    if obs_only and x_only and isinstance(obs_only, list):
        raise ValueError(
            "Can not use columns_obs_only together with columns_x_only with a single AnnData object. "
            "At least one has to be None!"
        )
    else:
        common_keys = obs_only.keys() & x_only.keys()  # type: ignore
        if common_keys:
            raise ValueError(
                "Can not use columns_obs_only together with columns_x_only for a single AnnData object. "
                "The following anndata identifiers where found"
                f"in both: {','.join(key for key in common_keys)}!"
            )


class IndexNotFoundError(Exception):
    pass


class CachingNotSupported(Exception):
    pass


class ExtensionMissingError(Exception):
    pass


class CacheExistsException(Exception):
    pass



import numpy as np
import pytest
from anndata import AnnData
from pandas import DataFrame

from ehrapy.preprocessing import summarize_measurements


@pytest.fixture
def adata_to_expand():
    np.random.seed(42)
    row_ids = ["pat1", "pat1", "pat1", "pat2", "pat2", "pat3"]
    measurement1 = np.random.choice([0, 1], size=6)
    measurement2 = np.random.uniform(0, 20, size=6)
    measurement3 = np.random.uniform(0, 20, size=6)
    data_dict = {"measurement1": measurement1, "measurement2": measurement2, "measurement3": measurement3}
    data_df = DataFrame(data_dict, index=row_ids)
    adata = AnnData(X=data_df)

    return adata


def test_all_statistics(adata_to_expand):
    transformed_adata = summarize_measurements(
        adata_to_expand,
    )

    assert transformed_adata.shape == (3, 9)  # (3 patients, 3 measurements * 3 statistics)
    assert np.allclose(
        transformed_adata[:, "measurement2_min"].X.reshape(-1), np.array([[3.1198905, 1.1616722, 12.0223]])
    )
    assert np.allclose(
        transformed_adata[:, "measurement2_max"].X.reshape(-1), np.array([[11.973169, 17.323523, 12.0223]])
    )
    assert np.allclose(
        transformed_adata[:, "measurement2_mean"].X.reshape(-1), np.array([[6.071144, 9.242598, 12.0223]])
    )


def test_var_names_subset(adata_to_expand):
    transformed_adata = summarize_measurements(
        adata_to_expand,
        var_names=["measurement1", "measurement2"],
    )

    assert transformed_adata.shape == (3, 6)  # (3 patients, 2 measurements * 3 statistics)


def test_statistics_subset(adata_to_expand):
    transformed_adata = summarize_measurements(adata_to_expand, statistics=["min"])

    assert transformed_adata.shape == (3, 3)  # (3 patients, 3 measurements * 1 statistics)



from __future__ import annotations

import copy
from pathlib import Path
from typing import TYPE_CHECKING, Literal

import numpy as np
import pandas as pd
from rich import print
from thefuzz import process

if TYPE_CHECKING:
    from collections.abc import Collection

    from anndata import AnnData


def qc_metrics(
    adata: AnnData, qc_vars: Collection[str] = (), layer: str = None, inplace: bool = True
) -> pd.DataFrame | None:
    """Calculates various quality control metrics.

    Uses the original values to calculate the metrics and not the encoded ones.
    Look at the return type for a more in depth description of the calculated metrics.

    Args:
        adata: Annotated data matrix.
        qc_vars: Optional List of vars to calculate additional metrics for.
        layer: Layer to use to calculate the metrics.
        inplace: Whether to add the metrics to obs/var or to solely return a Pandas DataFrame.

    Returns:
        Pandas DataFrame of all calculated QC metrics.

        Observation level metrics include:

        - `missing_values_abs`: Absolute amount of missing values.
        - `missing_values_pct`: Relative amount of missing values in percent.

        Feature level metrics include:

        - `missing_values_abs`: Absolute amount of missing values.
        - `missing_values_pct`: Relative amount of missing values in percent.
        - `mean`: Mean value of the features.
        - `median`: Median value of the features.
        - `std`: Standard deviation of the features.
        - `min`: Minimum value of the features.
        - `max`: Maximum value of the features.

        Examples:
            >>> import ehrapy as ep
            >>> import seaborn as sns
            >>> adata = ep.dt.mimic_2(encoded=True)
            >>> ep.pp.qc_metrics(adata)
            >>> sns.displot(adata.obs["missing_values_abs"])
    """
    obs_metrics = _obs_qc_metrics(adata, layer, qc_vars)
    var_metrics = _var_qc_metrics(adata, layer)

    if inplace:
        adata.obs[obs_metrics.columns] = obs_metrics
        adata.var[var_metrics.columns] = var_metrics

    return obs_metrics, var_metrics


def _missing_values(
    arr: np.ndarray, shape: tuple[int, int] = None, df_type: Literal["obs", "var"] = "obs"
) -> np.ndarray:
    """Calculates the absolute or relative amount of missing values.

    Args:
        arr: Numpy array containing a data row which is a subset of X (mtx).
        shape: Shape of X (mtx).
        df_type: Whether to calculate the proportions for obs or var. One of 'obs' or 'var'. Defaults to 'obs' .

    Returns:
        Absolute or relative amount of missing values.
    """
    if shape is None:
        return pd.isnull(arr).sum()
    else:
        n_rows, n_cols = shape
        if df_type == "obs":
            return (pd.isnull(arr).sum() / n_cols) * 100
        else:
            return (pd.isnull(arr).sum() / n_rows) * 100


def _obs_qc_metrics(
    adata: AnnData, layer: str = None, qc_vars: Collection[str] = (), log1p: bool = True
) -> pd.DataFrame:
    """Calculates quality control metrics for observations.

    See :func:`~ehrapy.preprocessing._quality_control.calculate_qc_metrics` for a list of calculated metrics.

    Args:
        adata: Annotated data matrix.
        layer: Layer containing the actual data matrix.
        qc_vars: A list of previously calculated QC metrics to calculate summary statistics for.
        log1p: Whether to apply log1p normalization for the QC metrics. Only used with parameter 'qc_vars'.

    Returns:
        A Pandas DataFrame with the calculated metrics.
    """
    obs_metrics = pd.DataFrame(index=adata.obs_names)
    var_metrics = pd.DataFrame(index=adata.var_names)
    mtx = adata.X if layer is None else adata.layers[layer]
    if "original_values_categoricals" in adata.uns:
        for original_values_categorical in list(adata.uns["original_values_categoricals"]):
            mtx = mtx.astype(object)
            index = np.where(var_metrics.index.str.contains(original_values_categorical))[0]
            mtx[:, index[0]] = np.squeeze(
                np.where(
                    adata.uns["original_values_categoricals"][original_values_categorical].astype(object) == "nan",
                    np.nan,
                    adata.uns["original_values_categoricals"][original_values_categorical].astype(object),
                )
            )

    obs_metrics["missing_values_abs"] = np.apply_along_axis(_missing_values, 1, mtx)
    obs_metrics["missing_values_pct"] = np.apply_along_axis(_missing_values, 1, mtx, shape=mtx.shape, df_type="obs")

    # Specific QC metrics
    for qc_var in qc_vars:
        obs_metrics[f"total_features_{qc_var}"] = np.ravel(mtx[:, adata.var[qc_var].values].sum(axis=1))
        if log1p:
            obs_metrics[f"log1p_total_features_{qc_var}"] = np.log1p(obs_metrics[f"total_features_{qc_var}"])
        obs_metrics["total_features"] = np.ravel(mtx.sum(axis=1))
        obs_metrics[f"pct_features_{qc_var}"] = (
            obs_metrics[f"total_features_{qc_var}"] / obs_metrics["total_features"] * 100
        )

    return obs_metrics


def _var_qc_metrics(adata: AnnData, layer: str = None) -> pd.DataFrame:
    """Calculates quality control metrics for features.

    See :func:`~ehrapy.preprocessing._quality_control.calculate_qc_metrics` for a list of calculated metrics.

    Args:
        adata: Annotated data matrix.
        layer: Layer containing the matrix to calculate the metrics for.

    Returns:
        Pandas DataFrame with the calculated metrics.
    """
    var_metrics = pd.DataFrame(index=adata.var_names)
    mtx = adata.X if layer is None else adata.layers[layer]
    categorical_indices = np.ndarray([0], dtype=int)
    if "original_values_categoricals" in adata.uns:
        for original_values_categorical in list(adata.uns["original_values_categoricals"]):
            mtx = copy.deepcopy(mtx.astype(object))
            index = np.where(var_metrics.index.str.startswith("ehrapycat_" + original_values_categorical))[0]
            mtx[:, index] = np.tile(
                np.where(
                    adata.uns["original_values_categoricals"][original_values_categorical].astype(object) == "nan",
                    np.nan,
                    adata.uns["original_values_categoricals"][original_values_categorical].astype(object),
                ),
                mtx[:, index].shape[1],
            )
            categorical_indices = np.concatenate([categorical_indices, index])
    non_categorical_indices = np.ones(mtx.shape[1], dtype=bool)
    non_categorical_indices[categorical_indices] = False
    var_metrics["missing_values_abs"] = np.apply_along_axis(_missing_values, 0, mtx)
    var_metrics["missing_values_pct"] = np.apply_along_axis(_missing_values, 0, mtx, shape=mtx.shape, df_type="var")

    var_metrics["mean"] = np.nan
    var_metrics["median"] = np.nan
    var_metrics["standard_deviation"] = np.nan
    var_metrics["min"] = np.nan
    var_metrics["max"] = np.nan

    try:
        var_metrics.loc[non_categorical_indices, "mean"] = np.nanmean(
            np.array(mtx[:, non_categorical_indices], dtype=np.float64), axis=0
        )
        var_metrics.loc[non_categorical_indices, "median"] = np.nanmedian(
            np.array(mtx[:, non_categorical_indices], dtype=np.float64), axis=0
        )
        var_metrics.loc[non_categorical_indices, "standard_deviation"] = np.nanstd(
            np.array(mtx[:, non_categorical_indices], dtype=np.float64), axis=0
        )
        var_metrics.loc[non_categorical_indices, "min"] = np.nanmin(
            np.array(mtx[:, non_categorical_indices], dtype=np.float64), axis=0
        )
        var_metrics.loc[non_categorical_indices, "max"] = np.nanmax(
            np.array(mtx[:, non_categorical_indices], dtype=np.float64), axis=0
        )
    except (TypeError, ValueError):
        print("[bold yellow]TypeError! Setting quality control metrics to nan. Did you encode your data?")

    return var_metrics


def qc_lab_measurements(
    adata: AnnData,
    reference_table: pd.DataFrame = None,
    measurements: list[str] = None,
    unit: Literal["traditional", "SI"] = None,
    layer: str = None,
    threshold: int = 20,
    age_col: str = None,
    age_range: str = None,
    sex_col: str = None,
    sex: str = None,
    ethnicity_col: str = None,
    ethnicity: str = None,
    copy: bool = False,
    verbose: bool = False,
) -> AnnData:
    """Examines lab measurements for reference ranges and outliers.

    Source:
        The used reference values were obtained from https://accessmedicine.mhmedical.com/content.aspx?bookid=1069&sectionid=60775149 .
        This table is compiled from data in the following sources:

        * Tietz NW, ed. Clinical Guide to Laboratory Tests. 3rd ed. Philadelphia: WB Saunders Co; 1995;
        * Laposata M. SI Unit Conversion Guide. Boston: NEJM Books; 1992;
        * American Medical Association Manual of Style: A Guide for Authors and Editors. 9th ed. Chicago: AMA; 1998:486–503. Copyright 1998, American Medical Association;
        * Jacobs DS, DeMott WR, Oxley DK, eds. Jacobs & DeMott Laboratory Test Handbook With Key Word Index. 5th ed. Hudson, OH: Lexi-Comp Inc; 2001;
        * Henry JB, ed. Clinical Diagnosis and Management by Laboratory Methods. 20th ed. Philadelphia: WB Saunders Co; 2001;
        * Kratz A, et al. Laboratory reference values. N Engl J Med. 2006;351:1548–1563; 7) Burtis CA, ed. Tietz Textbook of Clinical Chemistry and Molecular Diagnostics. 5th ed. St. Louis: Elsevier; 2012.

        This version of the table of reference ranges was reviewed and updated by Jessica Franco-Colon, PhD, and Kay Brooks.

    Limitations:
        * Reference ranges differ between continents, countries and even laboratories (https://informatics.bmj.com/content/28/1/e100419).
          The default values used here are only one of many options.
        * Ensure that the values used as input are provided with the correct units. We recommend the usage of SI values.
        * The reference values pertain to adults. Many of the reference ranges need to be adapted for children.
        * By default if no gender is provided and no unisex values are available, we use the **male** reference ranges.
        * The used reference ranges may be biased for ethnicity. Please examine the primary sources if required.
        * We recommend a glance at https://www.nature.com/articles/s41591-021-01468-6 for the effect of such covariates.

    Additional values:
        * Interleukin-6 based on https://pubmed.ncbi.nlm.nih.gov/33155686/

    If you want to specify your own table as a Pandas DataFrame please examine the existing default table.
    Ethnicity and age columns can be added.
    https://github.com/theislab/ehrapy/blob/main/ehrapy/preprocessing/laboratory_reference_tables/laposata.tsv

    Args:
        adata: Annotated data matrix.
        reference_table: A custom DataFrame with reference values. Defaults to the laposata table if not specified.
        measurements: A list of measurements to check.
        unit: The unit of the measurements. Defaults to 'traditional'.
        layer: Layer containing the matrix to calculate the metrics for.
        threshold: Minimum required matching confidence score of the fuzzysearch.
                   0 = no matches, 100 = all must match. Defaults to 20.
        age_col: Column containing age values.
        age_range: The inclusive age-range to filter for such as 5-99.
        sex_col: Column containing sex values. Column must contain 'U', 'M' or 'F'.
        sex: Sex to filter the reference values for. Use U for unisex which uses male values when male and female conflict.
             Defaults to 'U|M'.
        ethnicity_col: Column containing ethnicity values.
        ethnicity: Ethnicity to filter for.
        copy: Whether to return a copy. Defaults to False.
        verbose: Whether to have verbose stdout. Notifies user of matched columns and value ranges.

    Returns:
        A modified AnnData object (copy if specified).

    Examples:
        >>> import ehrapy as ep
        >>> adata = ep.dt.mimic_2(encoded=True)
        >>> ep.pp.qc_lab_measurements(adata, measurements=["potassium_first"], verbose=True)
    """
    if copy:
        adata = adata.copy()

    preprocessing_dir = Path(__file__).parent.resolve()
    if reference_table is None:
        reference_table = pd.read_csv(
            f"{preprocessing_dir}/laboratory_reference_tables/laposata.tsv", sep="\t", index_col="Measurement"
        )

    for measurement in measurements:
        best_column_match, score = process.extractOne(
            query=measurement, choices=reference_table.index, score_cutoff=threshold
        )
        if best_column_match is None:
            print(f"[bold yellow]Unable to find a match for {measurement}")
            continue
        if verbose:
            print(
                f"[bold blue]Detected [green]{best_column_match}[blue] for [green]{measurement}[blue] with score [green]{score}."
            )

        reference_column = "SI Reference Interval" if unit == "SI" else "Traditional Reference Interval"

        # Fetch all non None columns from the reference statistics
        not_none_columns = [col for col in [sex_col, age_col, ethnicity_col] if col is not None]
        not_none_columns.append(reference_column)
        reference_values = reference_table.loc[[best_column_match], not_none_columns]

        additional_columns = False
        if sex_col or age_col or ethnicity_col:  # check if additional columns were provided
            additional_columns = True

        # Check if multiple reference values occur and no additional information is available:
        if reference_values.shape[0] > 1 and additional_columns is False:
            raise ValueError(
                f"Several options for {best_column_match} reference value are available. Please specify sex, age or "
                f"ethnicity columns and their values."
            )

        try:
            if age_col:
                min_age, max_age = age_range.split("-")
                reference_values = reference_values[
                    (reference_values[age_col].str.split("-").str[0].astype(int) >= int(min_age))
                    and (reference_values[age_col].str.split("-").str[1].astype(int) <= int(max_age))
                ]
            if sex_col:
                sexes = "U|M" if sex is None else sex
                reference_values = reference_values[reference_values[sex_col].str.contains(sexes)]
            if ethnicity_col:
                reference_values = reference_values[reference_values[ethnicity_col].isin([ethnicity])]

            if layer is not None:
                actual_measurements = adata[:, measurement].layers[layer]
            else:
                actual_measurements = adata[:, measurement].X
        except TypeError:
            print(f"[bold yellow]Unable to find specified reference values for {measurement}.")

        check = reference_values[reference_column].values
        check_str: str = np.array2string(check)
        check_str = check_str.replace("[", "").replace("]", "").replace("'", "")
        if "<" in check_str:
            upperbound = float(check_str.replace("<", ""))
            if verbose:
                print(f"[bold blue]Using upperbound [green]{upperbound}")

            upperbound_check_results = actual_measurements < upperbound
            upperbound_check_results_array: np.ndarray = upperbound_check_results.copy()
            adata.obs[f"{measurement} normal"] = upperbound_check_results_array
        elif ">" in check_str:
            lower_bound = float(check_str.replace(">", ""))
            if verbose:
                print(f"[bold blue]Using lowerbound [green]{lower_bound}")

            lower_bound_check_results = actual_measurements > lower_bound
            lower_bound_check_results_array = lower_bound_check_results.copy()
            adata.obs[f"{measurement} normal"] = lower_bound_check_results_array
        else:  # "-" range case
            min_value = float(check_str.split("-")[0])
            max_value = float(check_str.split("-")[1])
            if verbose:
                print(f"[bold blue]Using minimum of [green]{min_value}[blue] and maximum of [green]{max_value}")

            range_check_results = (actual_measurements >= min_value) & (actual_measurements <= max_value)
            range_check_results_array: np.ndarray = range_check_results.copy()
            adata.obs[f"{measurement} normal"] = range_check_results_array

    if copy:
        return adata



import re
from collections.abc import Iterable, Iterator

from sphinx.application import Sphinx
from sphinx.ext.napoleon import NumpyDocstring


def _process_return(lines: Iterable[str]) -> Iterator[str]:
    for line in lines:
        m = re.fullmatch(r"(?P<param>\w+)\s+:\s+(?P<type>[\w.]+)", line)
        if m:
            # Once this is in scanpydoc, we can use the fancy hover stuff
            yield f'**{m["param"]}** : :class:`~{m["type"]}`'
        else:
            yield line


def _parse_returns_section(self: NumpyDocstring, section: str) -> list[str]:
    lines_raw = list(_process_return(self._dedent(self._consume_to_next_section())))
    lines: list[str] = self._format_block(":returns: ", lines_raw)
    if lines and lines[-1]:
        lines.append("")
    return lines


def setup(app: Sphinx) -> None:
    NumpyDocstring._parse_returns_section = _parse_returns_section



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from pandas import CategoricalDtype

from ehrapy.io._read import read_csv, read_fhir, read_h5ad

CURRENT_DIR = Path(__file__).parent
_TEST_PATH = f"{CURRENT_DIR}/test_data_io"
_TEST_PATH_H5AD = f"{CURRENT_DIR}/test_data_io/h5ad"
_TEST_PATH_MULTIPLE = f"{CURRENT_DIR}/test_data_io_multiple"
_TEST_PATH_FHIR = f"{CURRENT_DIR}/test_data_io/fhir/json"


class TestRead:
    def test_read_csv(self):
        adata = read_csv(dataset_path=f"{_TEST_PATH}/dataset_basic.csv")
        matrix = np.array(
            [[12, 14, 500, False], [13, 7, 330, False], [14, 10, 800, True], [15, 11, 765, True], [16, 3, 800, True]]
        )
        assert adata.X.shape == (5, 4)
        assert (adata.X == matrix).all()
        assert adata.var_names.to_list() == ["patient_id", "los_days", "b12_values", "survival"]
        assert (adata.layers["original"] == matrix).all()
        assert id(adata.layers["original"]) != id(adata.X)

    def test_read_tsv(self):
        adata = read_csv(dataset_path=f"{_TEST_PATH}/dataset_tsv.tsv", sep="\t")
        matrix = np.array(
            [
                [12, 54, 185.34, False],
                [13, 25, 175.39, True],
                [14, 36, 183.29, False],
                [15, 44, 173.93, True],
                [16, 27, 190.32, True],
            ]
        )
        assert adata.X.shape == (5, 4)
        assert (adata.X == matrix).all()
        assert adata.var_names.to_list() == ["patient_id", "age", "height", "gamer"]
        assert (adata.layers["original"] == matrix).all()
        assert id(adata.layers["original"]) != id(adata.X)

    def test_read_multiple_csv_to_anndatas(self):
        adatas = read_csv(dataset_path=f"{_TEST_PATH_MULTIPLE}")
        adata_ids = set(adatas.keys())
        assert all(adata_id in adata_ids for adata_id in {"dataset_non_num_with_missing", "dataset_num_with_missing"})
        assert set(adatas["dataset_non_num_with_missing"].var_names) == {
            "indexcol",
            "intcol",
            "strcol",
            "boolcol",
            "binary_col",
        }
        assert set(adatas["dataset_num_with_missing"].var_names) == {"col" + str(i) for i in range(1, 4)}

    def test_read_multiple_csvs_to_dfs(self):
        dfs = read_csv(dataset_path=f"{_TEST_PATH_MULTIPLE}", return_dfs=True)
        dfs_ids = set(dfs.keys())
        assert all(id in dfs_ids for id in {"dataset_non_num_with_missing", "dataset_num_with_missing"})
        assert set(dfs["dataset_non_num_with_missing"].columns) == {
            "indexcol",
            "intcol",
            "strcol",
            "boolcol",
            "binary_col",
            "datetime",
        }

    def test_read_multiple_csv_with_obs_only(self):
        adatas = read_csv(
            dataset_path=f"{_TEST_PATH_MULTIPLE}",
            columns_obs_only={"dataset_non_num_with_missing": ["strcol"], "dataset_num_with_missing": ["col1"]},
        )
        adata_ids = set(adatas.keys())
        assert all(adata_id in adata_ids for adata_id in {"dataset_non_num_with_missing", "dataset_num_with_missing"})
        assert set(adatas["dataset_non_num_with_missing"].var_names) == {"indexcol", "intcol", "boolcol", "binary_col"}
        assert set(adatas["dataset_num_with_missing"].var_names) == {"col" + str(i) for i in range(2, 4)}
        assert all(
            obs_name in set(adatas["dataset_non_num_with_missing"].obs.columns) for obs_name in {"datetime", "strcol"}
        )
        assert "col1" in set(adatas["dataset_num_with_missing"].obs.columns)

    def test_read_h5ad(self):
        adata = read_h5ad(dataset_path=f"{_TEST_PATH_H5AD}/dataset9.h5ad")

        assert adata.X.shape == (4, 3)
        assert set(adata.var_names) == {"col" + str(i) for i in range(1, 4)}
        assert set(adata.obs.columns) == set()

    def test_read_multiple_h5ad(self):
        adatas = read_h5ad(dataset_path=f"{_TEST_PATH_H5AD}")
        adata_ids = set(adatas.keys())

        assert all(adata_id in adata_ids for adata_id in {"dataset8", "dataset9"})
        assert set(adatas["dataset8"].var_names) == {"indexcol", "intcol", "boolcol", "binary_col", "strcol"}
        assert set(adatas["dataset9"].var_names) == {"col" + str(i) for i in range(1, 4)}
        assert all(obs_name in set(adatas["dataset8"].obs.columns) for obs_name in {"datetime"})

    def test_read_csv_without_index_column(self):
        adata = read_csv(dataset_path=f"{_TEST_PATH}/dataset_index.csv")
        matrix = np.array(
            [[1, 14, 500, False], [2, 7, 330, False], [3, 10, 800, True], [4, 11, 765, True], [5, 3, 800, True]]
        )
        assert adata.X.shape == (5, 4)
        assert (adata.X == matrix).all()
        assert adata.var_names.to_list() == ["clinic_id", "los_days", "b12_values", "survival"]
        assert (adata.layers["original"] == matrix).all()
        assert id(adata.layers["original"]) != id(adata.X)
        assert list(adata.obs.index) == ["0", "1", "2", "3", "4"]

    def test_read_csv_with_bools_obs_only(self):
        adata = read_csv(dataset_path=f"{_TEST_PATH}/dataset_basic.csv", columns_obs_only=["survival", "b12_values"])
        matrix = np.array([[12, 14], [13, 7], [14, 10], [15, 11], [16, 3]])
        assert adata.X.shape == (5, 2)
        assert (adata.X == matrix).all()
        assert adata.var_names.to_list() == ["patient_id", "los_days"]
        assert (adata.layers["original"] == matrix).all()
        assert id(adata.layers["original"]) != id(adata.X)
        assert set(adata.obs.columns) == {"b12_values", "survival"}
        assert pd.api.types.is_bool_dtype(adata.obs["survival"].dtype)
        assert pd.api.types.is_numeric_dtype(adata.obs["b12_values"].dtype)

    def test_read_csv_with_bools_and_cats_obs_only(self):
        adata = read_csv(
            dataset_path=f"{_TEST_PATH}/dataset_bools_and_str.csv", columns_obs_only=["b12_values", "name", "survival"]
        )
        matrix = np.array([[1, 14], [2, 7], [3, 10], [4, 11], [5, 3]])
        assert adata.X.shape == (5, 2)
        assert (adata.X == matrix).all()
        assert adata.var_names.to_list() == ["clinic_id", "los_days"]
        assert (adata.layers["original"] == matrix).all()
        assert id(adata.layers["original"]) != id(adata.X)
        assert set(adata.obs.columns) == {"b12_values", "survival", "name"}
        assert pd.api.types.is_bool_dtype(adata.obs["survival"].dtype)
        assert pd.api.types.is_numeric_dtype(adata.obs["b12_values"].dtype)
        assert isinstance(adata.obs["name"].dtype, CategoricalDtype)

    def test_set_default_index(self):
        adata = read_csv(dataset_path=f"{_TEST_PATH}/dataset_index.csv")
        assert adata.X.shape == (5, 4)
        assert not adata.obs_names.name
        assert list(adata.obs.index.values) == [f"{i}" for i in range(5)]

    def test_set_given_str_index(self):
        adata = read_csv(dataset_path=f"{_TEST_PATH}/dataset_basic.csv", index_column="los_days")
        assert adata.X.shape == (5, 3)
        assert adata.obs_names.name == "los_days"
        assert list(adata.obs.index.values) == ["14", "7", "10", "11", "3"]

    def test_set_given_int_index(self):
        adata = read_csv(dataset_path=f"{_TEST_PATH}/dataset_basic.csv", index_column=1)
        assert adata.X.shape == (5, 3)
        assert adata.obs_names.name == "los_days"
        assert list(adata.obs.index.values) == ["14", "7", "10", "11", "3"]

    def test_move_single_column_misspelled(self):
        with pytest.raises(ValueError):
            _ = read_csv(dataset_path=f"{_TEST_PATH}/dataset_basic.csv", columns_obs_only=["b11_values"])

    def test_move_single_column_to_obs(self):
        adata = read_csv(dataset_path=f"{_TEST_PATH}/dataset_basic.csv", columns_obs_only=["b12_values"])
        assert adata.X.shape == (5, 3)
        assert list(adata.obs.columns) == ["b12_values"]
        assert "b12_values" not in list(adata.var_names.values)

    def test_move_multiple_columns_to_obs(self):
        adata = read_csv(dataset_path=f"{_TEST_PATH}/dataset_basic.csv", columns_obs_only=["b12_values", "survival"])
        assert adata.X.shape == (5, 2)
        assert list(adata.obs.columns) == ["b12_values", "survival"]
        assert "b12_values" not in list(adata.var_names.values) and "survival" not in list(adata.var_names.values)

    def test_read_raises_error_with_duplicates_columns_only_single_1(self):
        with pytest.raises(ValueError):
            _ = read_csv(
                dataset_path=f"{_TEST_PATH}/dataset_basic.csv",
                columns_obs_only=["survival", "b12_values"],
                columns_x_only=["survival", "b12_values"],
            )

    def test_read_raises_error_with_duplicates_columns_only_single_2(self):
        with pytest.raises(ValueError):
            _ = read_csv(
                dataset_path=f"{_TEST_PATH}/dataset_basic.csv",
                columns_obs_only=["survival"],
                columns_x_only=["survival", "b12_values"],
            )

    def test_read_raises_error_with_duplicates_columns_only_multiple_1(self):
        with pytest.raises(ValueError):
            _ = read_csv(
                dataset_path=f"{_TEST_PATH_MULTIPLE}",
                columns_obs_only={
                    "dataset_non_num_with_missing": ["intcol"],
                    "dataset_num_with_missing": ["col1", "col2"],
                },
                columns_x_only={"dataset_non_num_with_missing": ["intcol"]},
            )

    def test_read_raises_error_with_duplicates_columns_only_multiple_2(self):
        with pytest.raises(ValueError):
            _ = read_csv(
                dataset_path=f"{_TEST_PATH_MULTIPLE}",
                columns_obs_only={
                    "dataset_non_num_with_missing": ["intcol"],
                    "dataset_num_with_missing": ["col1", "col2"],
                },
                columns_x_only={"dataset_non_num_with_missing": ["indexcol"], "dataset_num_with_missing": ["col3"]},
            )

    def test_move_single_column_to_x(self):
        adata = read_csv(dataset_path=f"{_TEST_PATH}/dataset_basic.csv", columns_x_only=["b12_values"])
        assert adata.X.shape == (5, 1)
        assert list(adata.var_names) == ["b12_values"]
        assert "b12_values" not in list(adata.obs.columns)
        assert all(obs_names in list(adata.obs.columns) for obs_names in ["los_days", "patient_id", "survival"])

    def test_move_multiple_columns_to_x(self):
        adata = read_csv(dataset_path=f"{_TEST_PATH}/dataset_basic.csv", columns_x_only=["b12_values", "survival"])
        assert adata.X.shape == (5, 2)
        assert all(var_names in list(adata.var_names) for var_names in ["b12_values", "survival"])
        assert all(obs_names in list(adata.obs.columns) for obs_names in ["los_days", "patient_id"])
        assert all(var_names not in list(adata.obs.columns) for var_names in ["b12_values", "survival"])

    def test_read_multiple_csv_with_x_only(self):
        adatas = read_csv(
            dataset_path=f"{_TEST_PATH_MULTIPLE}",
            columns_x_only={"dataset_non_num_with_missing": ["strcol"], "dataset_num_with_missing": ["col1"]},
        )
        adata_ids = set(adatas.keys())
        assert all(adata_id in adata_ids for adata_id in {"dataset_non_num_with_missing", "dataset_num_with_missing"})
        assert set(adatas["dataset_non_num_with_missing"].obs.columns) == {
            "indexcol",
            "intcol",
            "boolcol",
            "binary_col",
            "datetime",
        }
        assert set(adatas["dataset_num_with_missing"].obs.columns) == {"col" + str(i) for i in range(2, 4)}
        assert set(adatas["dataset_non_num_with_missing"].var_names) == {"strcol"}
        assert set(adatas["dataset_num_with_missing"].var_names) == {"col1"}

    def test_read_multiple_csv_with_x_only_2(self):
        adatas = read_csv(
            dataset_path=f"{_TEST_PATH_MULTIPLE}",
            columns_x_only={
                "dataset_non_num_with_missing": ["strcol", "intcol", "boolcol"],
                "dataset_num_with_missing": ["col1", "col3"],
            },
        )
        adata_ids = set(adatas.keys())
        assert all(adata_id in adata_ids for adata_id in {"dataset_non_num_with_missing", "dataset_num_with_missing"})
        assert set(adatas["dataset_non_num_with_missing"].obs.columns) == {"indexcol", "binary_col", "datetime"}
        assert set(adatas["dataset_num_with_missing"].obs.columns) == {"col2"}
        assert set(adatas["dataset_non_num_with_missing"].var_names) == {"strcol", "intcol", "boolcol"}
        assert set(adatas["dataset_num_with_missing"].var_names) == {"col1", "col3"}

    def test_read_fhir_json(self):
        adata = read_fhir(_TEST_PATH_FHIR)

        assert adata.shape == (4928, 80)
        assert "resource.birthDate" in adata.obs.columns

    def test_read_fhir_json_obs_only(self):
        adata = read_fhir(_TEST_PATH_FHIR, columns_obs_only=["fullUrl"])

        assert adata.shape == (4928, 79)
        assert "fullUrl" in adata.obs.columns



from pathlib import Path

import numpy as np
import pandas as pd
from anndata import AnnData

import ehrapy as ep

CURRENT_DIR = Path(__file__).parent
_TEST_PATH = f"{CURRENT_DIR}/test_data_nlp"


class TestMedCAT:
    def test_add_medcat_annotation_to_obs(self):
        # created manually a small dataset with annotations to use here
        adata = ep.io.read_csv(f"{_TEST_PATH}/dataset1.csv")
        adata.uns["medcat_annotations"] = pd.read_csv(f"{_TEST_PATH}/medcat_annotations1.csv")

        ep.tl.add_medcat_annotation_to_obs(adata, name="Diabetes")
        assert "Diabetes" in adata.obs.columns



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import ehrapy as ep
import ehrapy.tools.feature_ranking._rank_features_groups as _utils
from ehrapy.io._read import read_csv

CURRENT_DIR = Path(__file__).parent
_TEST_PATH = f"{CURRENT_DIR}/test_data_features_ranking"


class TestHelperFunctions:
    def test_adjust_pvalues(self):
        groups = ("group1", "group2")

        pvals: np.recarray = pd.DataFrame(
            {"group1": (0.01, 0.02, 0.03, 1.00), "group2": (0.04, 0.05, 0.06, 0.99)}
        ).to_records(index=False)
        expected_result_bh = pd.DataFrame(
            {"group1": (0.04, 0.04, 0.04, 1.00), "group2": (0.08, 0.08, 0.08, 0.99)}
        ).to_records(index=False)
        expected_result_bf = pd.DataFrame(
            {"group1": (0.04, 0.08, 0.12, 1.00), "group2": (0.16, 0.20, 0.24, 1.00)}
        ).to_records(index=False)

        result_bh = _utils._adjust_pvalues(pvals, corr_method="benjamini-hochberg")

        for group in groups:
            assert (pvals[group] <= result_bh[group]).all()
            assert np.allclose(result_bh[group], expected_result_bh[group])

        result_bf = _utils._adjust_pvalues(pvals, corr_method="bonferroni")

        for group in groups:
            assert (pvals[group] <= result_bf[group]).all()
            assert np.allclose(result_bf[group], expected_result_bf[group])

    def test_sort_features(self):
        def _is_sorted(arr: np.array):
            return np.all(arr[:-1] <= arr[1:])

        adata = ep.dt.mimic_2(encoded=False)
        adata.uns["rank_features_groups"] = {
            "params": {"whatever": "here is", "it does": "not matter", "but this key should be": "present for testing"},
            "names": pd.DataFrame(
                {
                    "group1": ("feature2", "feature1", "feature4", "feature3"),
                    "group2": ("feature5", "feature6", "feature7", "feature8"),
                }
            ).to_records(index=False),
            # Let's mix features in group1, and leave them sorted in group2
            "pvals": pd.DataFrame({"group1": (0.02, 0.01, 1.00, 0.03), "group2": (0.04, 0.05, 0.06, 0.99)}).to_records(
                index=False
            ),
            "scores": pd.DataFrame({"group1": (2, 1, 100, 3), "group2": (4, 5, 6, 7)}).to_records(index=False),
            "log2foldchanges": pd.DataFrame({"group1": (2, 1, 10, 3), "group2": (4, 5, 6, 7)}).to_records(index=False),
        }
        # Doesn't really matter that they are the same here but order should be preserved
        adata.uns["rank_features_groups"]["pvals_adj"] = adata.uns["rank_features_groups"]["pvals"].copy()

        _utils._sort_features(adata)

        # Check that every feature is sorted
        # Actually, some of them would be sorted in the opposite direction (e.g. scores) for real data
        # But what matters here is that the permutation is the same for every key and is based on adjusted p-values
        for key in adata.uns["rank_features_groups"]:
            if key == "params":
                continue
            assert _is_sorted(adata.uns["rank_features_groups"][key]["group1"])
            assert _is_sorted(adata.uns["rank_features_groups"][key]["group2"])

    def test_save_rank_features_result(self):
        groups = ("group1", "group2")

        adata = ep.dt.mimic_2(encoded=False)
        adata.uns["rank_features_groups"] = {
            "params": {"whatever": "here is", "it does": "not matter", "but this key should be": "present for testing"}
        }

        names = pd.DataFrame(
            {
                "group1": ("feature2", "feature1", "feature4", "feature3"),
                "group2": ("feature5", "feature6", "feature7", "feature8"),
            }
        ).to_records(index=False)
        pvals = pd.DataFrame({"group1": (0.02, 0.01, 1.00, 0.03), "group2": (0.04, 0.05, 0.06, 0.99)}).to_records(
            index=False
        )
        scores = pd.DataFrame({"group1": (2, 1, 100, 3), "group2": (4, 5, 6, 7)}).to_records(index=False)
        logfoldchanges = pd.DataFrame({"group1": (2, 1, 10, 3), "group2": (4, 5, 6, 7)}).to_records(index=False)

        # Check that adding only required keys works
        adata_only_required = adata.copy()
        _utils._save_rank_features_result(
            adata_only_required,
            key_added="rank_features_groups",
            groups_order=groups,
            names=names,
            scores=scores,
            pvals=pvals,
        )

        assert "names" in adata_only_required.uns["rank_features_groups"]
        assert "pvals" in adata_only_required.uns["rank_features_groups"]
        assert "scores" in adata_only_required.uns["rank_features_groups"]
        assert "log2foldchanges" not in adata_only_required.uns["rank_features_groups"]
        assert "pvals_adj" not in adata_only_required.uns["rank_features_groups"]
        assert "pts" not in adata_only_required.uns["rank_features_groups"]
        assert adata_only_required.uns["rank_features_groups"]["names"].dtype.names == groups
        assert (
            len(adata_only_required.uns["rank_features_groups"]["names"]) == 4
        )  # It only captures the length of each group
        assert len(adata_only_required.uns["rank_features_groups"]["pvals"]) == 4
        assert len(adata_only_required.uns["rank_features_groups"]["scores"]) == 4

        # Check that running the function on adata with existing keys merges the arrays correctly
        _utils._save_rank_features_result(
            adata_only_required,
            key_added="rank_features_groups",
            groups_order=groups,
            names=names,
            scores=scores,
            pvals=pvals,
        )
        assert len(adata_only_required.uns["rank_features_groups"]["names"]) == 8
        assert len(adata_only_required.uns["rank_features_groups"]["pvals"]) == 8
        assert len(adata_only_required.uns["rank_features_groups"]["scores"]) == 8

        # Check that adding other keys works
        adata_all_keys = adata.copy()
        _utils._save_rank_features_result(
            adata_all_keys,
            key_added="rank_features_groups",
            groups_order=groups,
            names=names,
            scores=scores,
            pvals=pvals,
            pvals_adj=pvals.copy(),
            pts=pvals.copy(),
            logfoldchanges=logfoldchanges,
        )

        assert "names" in adata_all_keys.uns["rank_features_groups"]
        assert "pvals" in adata_all_keys.uns["rank_features_groups"]
        assert "scores" in adata_all_keys.uns["rank_features_groups"]
        assert "logfoldchanges" in adata_all_keys.uns["rank_features_groups"]
        assert "pvals_adj" in adata_all_keys.uns["rank_features_groups"]
        assert "pts" in adata_all_keys.uns["rank_features_groups"]
        assert adata_all_keys.uns["rank_features_groups"]["names"].dtype.names == groups
        assert len(adata_all_keys.uns["rank_features_groups"]["names"]) == 4
        assert len(adata_all_keys.uns["rank_features_groups"]["pvals"]) == 4
        assert len(adata_all_keys.uns["rank_features_groups"]["pvals_adj"]) == 4
        assert len(adata_all_keys.uns["rank_features_groups"]["logfoldchanges"]) == 4
        assert len(adata_all_keys.uns["rank_features_groups"]["scores"]) == 4
        assert len(adata_all_keys.uns["rank_features_groups"]["pts"]) == 4

        # Check that passing empty objects doesn't add keys
        _utils._save_rank_features_result(
            adata,
            key_added="rank_features_groups",
            groups_order=groups,
            names=names,
            scores=scores,
            pvals=pvals,
            pvals_adj=[],
            pts=np.array([]),
            logfoldchanges=pd.DataFrame([]),
        )
        assert "names" in adata.uns["rank_features_groups"]
        assert "pvals" in adata.uns["rank_features_groups"]
        assert "scores" in adata.uns["rank_features_groups"]
        assert "logfoldchanges" not in adata.uns["rank_features_groups"]
        assert "pvals_adj" not in adata.uns["rank_features_groups"]
        assert "pts" not in adata.uns["rank_features_groups"]

    def test_get_groups_order(self):
        assert _utils._get_groups_order(groups_subset="all", group_names=("A", "B", "C"), reference="B") == (
            "A",
            "B",
            "C",
        )
        assert _utils._get_groups_order(groups_subset=("A", "B"), group_names=("A", "B", "C"), reference="B") == (
            "A",
            "B",
        )
        assert _utils._get_groups_order(groups_subset=("A", "B"), group_names=("A", "B", "C"), reference="rest") == (
            "A",
            "B",
        )
        assert _utils._get_groups_order(groups_subset=("A", "B"), group_names=("A", "B", "C"), reference="C") == (
            "A",
            "B",
            "C",
        )

        # Check that array with ints (e.g. leiden clusters) works correctly
        assert _utils._get_groups_order(groups_subset=(1, 2), group_names=np.arange(3), reference="rest") == ("1", "2")

        with pytest.raises(ValueError):
            # Reference not in group_names
            _utils._get_groups_order(groups_subset=("A", "B"), group_names=("A", "B", "C"), reference="D")

    def test_evaluate_categorical_features(self):
        adata = ep.dt.mimic_2(encoded=False)
        adata = ep.pp.encode(adata, autodetect=True, encodings="label")

        group_names = pd.Categorical(adata.obs["service_unit"].astype(str)).categories.tolist()

        for method in ("chi-square", "g-test", "freeman-tukey", "mod-log-likelihood", "neyman", "cressie-read"):
            names, scores, pvals, logfc, pts = _utils._evaluate_categorical_features(
                adata, groupby="service_unit", group_names=group_names, categorical_method=method
            )

            # Check that important fields are not empty
            assert len(names)
            assert len(scores)
            assert len(pvals)
            assert len(logfc)
            assert not len(pts)  # Because pts == False by default

            assert (pvals >= 0).all()
            assert (pvals <= 1).all()

            # Check that grouping feature is not in the results
            assert "service_unit" not in names
            assert "ehrapycat_service_unit" not in names

            # Check that the only other categorical feature is in the results
            assert "ehrapycat_day_icu_intime" in names


class TestRankFeaturesGroups:
    def test_real_dataset(self):
        adata = ep.dt.mimic_2(encoded=True)

        ep.tl.rank_features_groups(adata, groupby="service_unit")

        assert "rank_features_groups" in adata.uns
        assert "names" in adata.uns["rank_features_groups"]
        assert "pvals" in adata.uns["rank_features_groups"]
        assert "scores" in adata.uns["rank_features_groups"]
        assert "logfoldchanges" in adata.uns["rank_features_groups"]
        assert "pvals_adj" in adata.uns["rank_features_groups"]

        assert "params" in adata.uns["rank_features_groups"]
        assert "method" in adata.uns["rank_features_groups"]["params"]
        assert "categorical_method" in adata.uns["rank_features_groups"]["params"]
        assert "reference" in adata.uns["rank_features_groups"]["params"]
        assert "groupby" in adata.uns["rank_features_groups"]["params"]
        assert "layer" in adata.uns["rank_features_groups"]["params"]
        assert "corr_method" in adata.uns["rank_features_groups"]["params"]

    def test_only_continous_features(self):
        adata = ep.dt.mimic_2(encoded=True)
        adata.uns["non_numerical_columns"] = []

        ep.tl.rank_features_groups(adata, groupby="service_unit")
        assert "rank_features_groups" in adata.uns
        assert "names" in adata.uns["rank_features_groups"]
        assert "pvals" in adata.uns["rank_features_groups"]
        assert "scores" in adata.uns["rank_features_groups"]
        assert "logfoldchanges" in adata.uns["rank_features_groups"]
        assert "pvals_adj" in adata.uns["rank_features_groups"]

    def test_only_cat_features(self):
        adata = ep.dt.mimic_2(encoded=True)
        adata.uns["numerical_columns"] = []

        ep.tl.rank_features_groups(adata, groupby="service_unit")
        assert "rank_features_groups" in adata.uns
        assert "names" in adata.uns["rank_features_groups"]
        assert "pvals" in adata.uns["rank_features_groups"]
        assert "scores" in adata.uns["rank_features_groups"]
        assert "logfoldchanges" in adata.uns["rank_features_groups"]
        assert "pvals_adj" in adata.uns["rank_features_groups"]

    @pytest.mark.parametrize("field_to_rank", ["layer", "obs", "layer_and_obs"])
    def test_rank_adata_immutability_property(self, field_to_rank):
        """
        Test that rank_features_group does not modify the adata object passed to it,
        except for the desired .uns field.
        This test is important because to save memory, copies are made conservatively in rank_features_groups
        """
        adata = read_csv(
            dataset_path=f"{_TEST_PATH}/dataset1.csv", columns_x_only=["station", "sys_bp_entry", "dia_bp_entry"]
        )
        adata = ep.pp.encode(adata, encodings={"label": ["station"]})
        adata_orig = adata.copy()

        ep.tl.rank_features_groups(adata, groupby="disease", field_to_rank=field_to_rank)

        assert adata_orig.shape == adata.shape
        assert adata_orig.X.shape == adata.X.shape
        assert adata_orig.obs.shape == adata.obs.shape
        assert adata_orig.var.shape == adata.var.shape

        assert np.allclose(adata_orig.X, adata.X)
        assert np.array_equal(adata_orig.obs, adata.obs)

        assert "rank_features_groups" in adata.uns

    @pytest.mark.parametrize("field_to_rank", ["layer", "obs", "layer_and_obs"])
    def test_rank_features_groups_generates_outputs(self, field_to_rank):
        """
        Test that the desired output is generated
        """

        adata = read_csv(
            dataset_path=f"{_TEST_PATH}/dataset1.csv",
            columns_obs_only=["disease", "station", "sys_bp_entry", "dia_bp_entry"],
        )

        ep.tl.rank_features_groups(adata, groupby="disease", field_to_rank=field_to_rank)

        # check standard rank_features_groups entries
        assert "names" in adata.uns["rank_features_groups"]
        assert "pvals" in adata.uns["rank_features_groups"]
        assert "scores" in adata.uns["rank_features_groups"]
        assert "pvals_adj" in adata.uns["rank_features_groups"]
        assert "logfoldchanges" in adata.uns["rank_features_groups"]
        assert "log2foldchanges" not in adata.uns["rank_features_groups"]
        assert "pts" not in adata.uns["rank_features_groups"]

        if field_to_rank == "layer" or field_to_rank == "obs":
            assert len(adata.uns["rank_features_groups"]["names"]) == 3  # It only captures the length of each group
            assert len(adata.uns["rank_features_groups"]["pvals"]) == 3
            assert len(adata.uns["rank_features_groups"]["scores"]) == 3

        elif field_to_rank == "layer_and_obs":
            assert len(adata.uns["rank_features_groups"]["names"]) == 6  # It only captures the length of each group
            assert len(adata.uns["rank_features_groups"]["pvals"]) == 6
            assert len(adata.uns["rank_features_groups"]["scores"]) == 6

    def test_rank_features_groups_consistent_results(self):
        adata_features_in_x = read_csv(
            dataset_path=f"{_TEST_PATH}/dataset1.csv",
            columns_x_only=["station", "sys_bp_entry", "dia_bp_entry", "glucose"],
        )
        adata_features_in_x = ep.pp.encode(adata_features_in_x, encodings={"label": ["station"]})

        adata_features_in_obs = read_csv(
            dataset_path=f"{_TEST_PATH}/dataset1.csv",
            columns_obs_only=["disease", "station", "sys_bp_entry", "dia_bp_entry", "glucose"],
        )

        adata_features_in_x_and_obs = read_csv(
            dataset_path=f"{_TEST_PATH}/dataset1.csv",
            columns_obs_only=["disease", "station"],
        )
        # to keep the same variables as in the datsets above, in order to make the comparison of consistency
        adata_features_in_x_and_obs = adata_features_in_x_and_obs[:, ["sys_bp_entry", "dia_bp_entry", "glucose"]]

        ep.tl.rank_features_groups(adata_features_in_x, groupby="disease")
        ep.tl.rank_features_groups(adata_features_in_obs, groupby="disease", field_to_rank="obs")
        ep.tl.rank_features_groups(adata_features_in_x_and_obs, groupby="disease", field_to_rank="layer_and_obs")

        for record in adata_features_in_x.uns["rank_features_groups"]["names"].dtype.names:
            assert np.allclose(
                adata_features_in_x.uns["rank_features_groups"]["scores"][record],
                adata_features_in_obs.uns["rank_features_groups"]["scores"][record],
            )
            assert np.allclose(
                np.array(adata_features_in_x.uns["rank_features_groups"]["pvals"][record]),
                np.array(adata_features_in_obs.uns["rank_features_groups"]["pvals"][record]),
            )
            assert np.array_equal(
                np.array(adata_features_in_x.uns["rank_features_groups"]["names"][record]),
                np.array(adata_features_in_obs.uns["rank_features_groups"]["names"][record]),
            )
        for record in adata_features_in_x.uns["rank_features_groups"]["names"].dtype.names:
            assert np.allclose(
                adata_features_in_x.uns["rank_features_groups"]["scores"][record],
                adata_features_in_x_and_obs.uns["rank_features_groups"]["scores"][record],
            )
            assert np.allclose(
                np.array(adata_features_in_x.uns["rank_features_groups"]["pvals"][record]),
                np.array(adata_features_in_x_and_obs.uns["rank_features_groups"]["pvals"][record]),
            )
            assert np.array_equal(
                np.array(adata_features_in_x.uns["rank_features_groups"]["names"][record]),
                np.array(adata_features_in_x_and_obs.uns["rank_features_groups"]["names"][record]),
            )

    def test_rank_features_group_column_to_rank(self):
        adata = read_csv(
            dataset_path=f"{_TEST_PATH}/dataset1.csv",
            columns_obs_only=["disease", "station", "sys_bp_entry", "dia_bp_entry"],
            index_column="idx",
        )

        # get a fresh adata for every test to not have any side effects
        adata_copy = adata.copy()

        ep.tl.rank_features_groups(adata, groupby="disease", columns_to_rank="all")
        assert len(adata.uns["rank_features_groups"]["names"]) == 2

        # want to check a "complete selection" works
        adata = adata_copy.copy()
        ep.tl.rank_features_groups(adata, groupby="disease", columns_to_rank={"var_names": ["glucose", "weight"]})
        assert len(adata.uns["rank_features_groups"]["names"]) == 2

        # want to check a "sub-selection" works
        adata = adata_copy.copy()
        ep.tl.rank_features_groups(adata, groupby="disease", columns_to_rank={"var_names": ["glucose"]})
        assert len(adata.uns["rank_features_groups"]["names"]) == 1

        # want to check a "complete" selection works
        adata = adata_copy.copy()
        ep.tl.rank_features_groups(
            adata,
            groupby="disease",
            field_to_rank="obs",
            columns_to_rank={"obs_names": ["station", "sys_bp_entry", "dia_bp_entry"]},
        )
        assert len(adata.uns["rank_features_groups"]["names"]) == 3

        # want to check a "sub-selection" selection works
        adata = adata_copy.copy()
        ep.tl.rank_features_groups(
            adata,
            groupby="disease",
            field_to_rank="obs",
            columns_to_rank={"obs_names": ["sys_bp_entry", "dia_bp_entry"]},
        )
        assert len(adata.uns["rank_features_groups"]["names"]) == 2



# Typing Column
# -----------------------
# The column name and used values in adata.var for column types.

EHRAPY_TYPE_KEY = "ehrapy_column_type"
NUMERIC_TAG = "numeric"
NON_NUMERIC_TAG = "non_numeric"
NON_NUMERIC_ENCODED_TAG = "non_numeric_encoded"



from __future__ import annotations

import sys
from datetime import datetime

import session_info
from rich import print

from ehrapy import __version__
from ehrapy.logging import _versions_dependencies


def print_versions():  # pragma: no cover
    """Print print versions of imported packages.

    Examples:
        >>> import ehrapy as ep
        >>> ep.print_versions()
    """
    try:
        session_info.show(
            dependencies=True,
            html=False,
            excludes=[
                "builtins",
                "stdlib_list",
                "importlib_metadata",
                "jupyter_core"
                # Special module present if test coverage being calculated
                # https://gitlab.com/joelostblom/session_info/-/issues/10
                "$coverage",
            ],
        )
    except AttributeError:
        print("[bold yellow]Unable to fetch versions for one or more dependencies.")
        pass


def print_version_and_date(*, file=None):  # pragma: no cover
    """Useful for starting a notebook so you see when you started working."""
    if file is None:
        file = sys.stdout
    print(
        f"Running ehrapy {__version__}, " f"on {datetime.now():%Y-%m-%d %H:%M}.",
        file=file,
    )


def print_header(*, file=None):  # pragma: no cover
    """Versions that might influence the numerical results.

    Matplotlib and Seaborn are excluded from this.
    """
    _DEPENDENCIES_NUMERICS = [
        "scanpy",
        "anndata",
        "umap",
        "numpy",
        "scipy",
        "pandas",
        ("sklearn", "scikit-learn"),
        "statsmodels",
        ("igraph", "python-igraph"),
        "leidenalg",
        "pynndescent",
    ]

    modules = ["ehrapy"] + _DEPENDENCIES_NUMERICS
    print(
        " ".join(f"{mod}=={ver}" for mod, ver in _versions_dependencies(modules)),
        file=file or sys.stdout,
    )



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from anndata import AnnData

import ehrapy as ep
from ehrapy.io._read import read_csv
from ehrapy.preprocessing._encode import encode
from ehrapy.preprocessing._quality_control import _obs_qc_metrics, _var_qc_metrics

CURRENT_DIR = Path(__file__).parent
_TEST_PATH = f"{CURRENT_DIR}/test_preprocessing"
_TEST_PATH_ENCODE = f"{CURRENT_DIR}/test_data_encode"


@pytest.fixture
def obs_data():
    return {
        "disease": ["cancer", "tumor"],
        "country": ["Germany", "switzerland"],
        "sex": ["male", "female"],
    }


@pytest.fixture
def var_data():
    return {
        "alive": ["yes", "no", "maybe"],
        "hospital": ["hospital 1", "hospital 2", "hospital 1"],
        "crazy": ["yes", "yes", "yes"],
    }


@pytest.fixture
def missing_values_adata(obs_data, var_data):
    return AnnData(
        X=np.array([[0.21, np.nan, 41.42], [np.nan, np.nan, 7.234]], dtype=np.float32),
        obs=pd.DataFrame(data=obs_data),
        var=pd.DataFrame(data=var_data, index=["Acetaminophen", "hospital", "crazy"]),
    )


@pytest.fixture
def lab_measurements_simple_adata(obs_data, var_data):
    X = np.array([[73, 0.02, 1.00], [148, 0.25, 3.55]], dtype=np.float32)
    return AnnData(
        X=X,
        obs=pd.DataFrame(data=obs_data),
        var=pd.DataFrame(data=var_data, index=["Acetaminophen", "Acetoacetic acid", "Beryllium, toxic"]),
    )


@pytest.fixture
def lab_measurements_layer_adata(obs_data, var_data):
    X = np.array([[73, 0.02, 1.00], [148, 0.25, 3.55]], dtype=np.float32)
    return AnnData(
        X=X,
        obs=pd.DataFrame(data=obs_data),
        var=pd.DataFrame(data=var_data, index=["Acetaminophen", "Acetoacetic acid", "Beryllium, toxic"]),
        layers={"layer_copy": X},
    )


def test_obs_qc_metrics(missing_values_adata):
    obs_metrics = _obs_qc_metrics(missing_values_adata)

    assert np.array_equal(obs_metrics["missing_values_abs"].values, np.array([1, 2]))
    assert np.allclose(obs_metrics["missing_values_pct"].values, np.array([33.3333, 66.6667]))


def test_var_qc_metrics(missing_values_adata):
    var_metrics = _var_qc_metrics(missing_values_adata)

    assert np.array_equal(var_metrics["missing_values_abs"].values, np.array([1, 2, 0]))
    assert np.allclose(var_metrics["missing_values_pct"].values, np.array([50.0, 100.0, 0.0]))
    assert np.allclose(var_metrics["mean"].values, np.array([0.21, np.nan, 24.327]), equal_nan=True)
    assert np.allclose(var_metrics["median"].values, np.array([0.21, np.nan, 24.327]), equal_nan=True)
    assert np.allclose(var_metrics["min"].values, np.array([0.21, np.nan, 7.234]), equal_nan=True)
    assert np.allclose(var_metrics["max"].values, np.array([0.21, np.nan, 41.419998]), equal_nan=True)


def test_obs_nan_qc_metrics():
    adata = read_csv(dataset_path=f"{_TEST_PATH_ENCODE}/dataset1.csv")
    adata.X[0][4] = np.nan
    adata2 = encode(adata, encodings={"one-hot": ["clinic_day"]})
    obs_metrics = _obs_qc_metrics(adata2)
    assert obs_metrics.iloc[0].iloc[0] == 1


def test_var_nan_qc_metrics():
    adata = read_csv(dataset_path=f"{_TEST_PATH_ENCODE}/dataset1.csv")
    adata.X[0][4] = np.nan
    adata2 = encode(adata, encodings={"one-hot": ["clinic_day"]})
    var_metrics = _var_qc_metrics(adata2)
    assert var_metrics.iloc[0].iloc[0] == 1
    assert var_metrics.iloc[1].iloc[0] == 1
    assert var_metrics.iloc[2].iloc[0] == 1
    assert var_metrics.iloc[3].iloc[0] == 1
    assert var_metrics.iloc[4].iloc[0] == 1


def test_calculate_qc_metrics(missing_values_adata):
    obs_metrics, var_metrics = ep.pp.qc_metrics(missing_values_adata, inplace=True)

    assert obs_metrics is not None
    assert var_metrics is not None

    assert missing_values_adata.obs.missing_values_abs is not None
    assert missing_values_adata.var.missing_values_abs is not None


def test_qc_lab_measurements_simple(lab_measurements_simple_adata):
    expected_obs_data = pd.Series(
        data={
            "Acetaminophen normal": [True, True],
            "Acetoacetic acid normal": [True, False],
            "Beryllium, toxic normal": [False, True],
        }
    )

    ep.pp.qc_lab_measurements(
        lab_measurements_simple_adata,
        measurements=list(lab_measurements_simple_adata.var_names),
        unit="SI",
    )

    assert (
        list(lab_measurements_simple_adata.obs["Acetaminophen normal"]) == (expected_obs_data["Acetaminophen normal"])
    )
    assert (
        list(lab_measurements_simple_adata.obs["Acetoacetic acid normal"])
        == (expected_obs_data["Acetoacetic acid normal"])
    )
    assert (
        list(lab_measurements_simple_adata.obs["Beryllium, toxic normal"])
        == (expected_obs_data["Beryllium, toxic normal"])
    )


def test_qc_lab_measurements_simple_layer(lab_measurements_layer_adata):
    expected_obs_data = pd.Series(
        data={
            "Acetaminophen normal": [True, True],
            "Acetoacetic acid normal": [True, False],
            "Beryllium, toxic normal": [False, True],
        }
    )

    ep.pp.qc_lab_measurements(
        lab_measurements_layer_adata,
        measurements=list(lab_measurements_layer_adata.var_names),
        unit="SI",
        layer="layer_copy",
    )

    assert list(lab_measurements_layer_adata.obs["Acetaminophen normal"]) == (expected_obs_data["Acetaminophen normal"])
    assert (
        list(lab_measurements_layer_adata.obs["Acetoacetic acid normal"])
        == (expected_obs_data["Acetoacetic acid normal"])
    )
    assert (
        list(lab_measurements_layer_adata.obs["Beryllium, toxic normal"])
        == (expected_obs_data["Beryllium, toxic normal"])
    )


def test_qc_lab_measurements_age():
    # TODO
    pass


def test_qc_lab_measurements_sex():
    # TODO
    pass


def test_qc_lab_measurements_ethnicity():
    # TODO
    pass


def test_qc_lab_measurements_multiple_measurements():
    data = pd.DataFrame(np.array([[100, 98], [162, 107]]), columns=["oxygen saturation", "glucose"], index=[0, 1])

    with pytest.raises(ValueError):
        adata = ep.ad.df_to_anndata(data)
        ep.pp.qc_lab_measurements(adata, measurements=["oxygen saturation", "glucose"], unit="SI")



from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd
from thefuzz import process

if TYPE_CHECKING:
    from collections.abc import Iterable

    from anndata import AnnData

    try:
        from medcat.cat import CAT

    except ModuleNotFoundError:
        pass


def _format_df_column(df: pd.DataFrame, column_name: str) -> list[tuple[int, str]]:
    """Format the df to match: formatted_data = [(row_id, row_text), (row_id, row_text), ...]
    as this is required by MedCAT's multiprocessing annotation step

    """
    formatted_data = []
    for id, row in df.iterrows():
        text = row[column_name]
        formatted_data.append((id, text))
    return formatted_data


def _flatten_annotated_results(annotation_results: dict) -> dict:
    """Flattens the nested set (usually 5 level nested) of annotation results.

    annotation_results is just a simple flattened dict with infos on all entities found
    """
    flattened_annotated_dict = {}
    entry_nr = 0

    # row numbers where the text column is located in the original data
    for row_id in annotation_results.keys():
        # all entities extracted from a given row
        entities = annotation_results[row_id]["entities"]
        for entity_id in entities.keys():
            # tokens are currently ignored, as they will not appear with the current basic model used by ehrapy from MedCAT
            if entity_id != "tokens":
                single_entity = {"row_nr": row_id}
                entity = entities[entity_id]
                # iterate over all info attributes of a single entity found in a specific row
                for entity_key in entity.keys():
                    if entity_key in ["pretty_name", "cui", "type_ids", "types"]:
                        single_entity[entity_key] = entities[entity_id][entity_key]
                    elif entity_key == "meta_anns":
                        single_entity[entity_key] = entities[entity_id][entity_key]["Status"]["value"]
                flattened_annotated_dict[entry_nr] = single_entity
                entry_nr += 1
    return flattened_annotated_dict


def annotate_text(
    adata: AnnData,
    cat: CAT,
    text_column: str,
    key_added: str = "medcat_annotations",
    n_proc: int = 2,
    batch_size_chars: int = 500000,
    copy: bool = False,
) -> AnnData | None:
    """Annotate the original free text data. Note this will only annotate non null rows.

    The result is a DataFrame.
    This DataFrame serves as the base for all further analyses, for example coloring UMAPs by specific diseases.

    Args:
        adata: AnnData object that holds the data to annotate.
        cat: MedCAT object.
        text_column: Name of the column that should be annotated.
        key_added: Key to add to adata.uns for the annotated results.
        n_proc: Number of processors to use.
        batch_size_chars: batch size to use for CAT's multiprocessing method.
        copy: Whether to copy adata or not.

    Returns:
        Returns `None` if `copy=False`, else returns an `AnnData` object. Sets the following fields;

        `adata.uns[key_added]` : :class:`pandas.DataFrame`
            DataFrame with the annotated results.

    """
    if copy:
        adata = adata.copy()
    non_null_text = pd.DataFrame(adata.obs[text_column][~adata.obs[text_column].isnull()])
    formatted_text_column = _format_df_column(non_null_text, text_column)
    results = cat.multiprocessing(formatted_text_column, batch_size_chars=batch_size_chars, nproc=n_proc)
    flattened_res = _flatten_annotated_results(results)

    # flatten annotated results into a Pandas DataFrame and remove duplicate entries; for example when a single entity like a disease is mentioned multiple times without any meaningful context changes
    # Example: The patient suffers from Diabetes. Cause of the Diabetes, he receives drug X.
    flattened_res_df = pd.DataFrame.from_dict(flattened_res, orient="index").drop_duplicates(
        subset=["cui", "row_nr", "meta_anns"]
    )

    # sort for row number in ascending order and reset index to keep index updated
    adata.uns[key_added] = flattened_res_df.sort_values(by=["row_nr"]).reset_index(drop=True)

    return adata if copy else None


def _filter_df_by_status(df: pd.DataFrame, status: str) -> pd.DataFrame:
    """Util function to filter passed dataframe by status."""
    df_res = df
    if status != "Both":
        if status not in {"Affirmed", "Other"}:
            raise StatusNotSupportedError(f"{status} is not available. Please use either Affirmed, Other or Both!")
        mask = df["meta_anns"].values == status
        df_res = df[mask]
    return df_res


def get_medcat_annotation_overview(
    adata: AnnData, status: str = "Affirmed", use_key: str = "medcat_annotations"
) -> pd.DataFrame:
    """Provide an overview for the annotation results. An overview will look like the following:

    cui (the CUI), nsubjects (from how many rows this one got extracted), type_ids (TUIs), name(name of the entitiy), perc_subjects (how many rows relative
    to absolute number of rows)

    Args:
        medcat_obj: The current MedCAT object which holds all infos on NLP analysis with MedCAT and ehrapy.
        n: Basically the parameter for head() of pandas Dataframe. How many of the most common entities should be shown?
        status: One of "Affirmed" (default), "Other" or "Both". Displays stats for either only affirmed entities, negated ones or both.
        use_key: Key to use for the annotated results.

    Returns:
        A Pandas DataFrame with the overview stats.
    """
    df = _filter_df_by_status(adata.uns[use_key], status)
    # group by CUI as this is a unique identifier per entity
    grouped = df.groupby("cui")
    # get absolute number of rows with this entity
    # note for overview, only one TUI and type is shown (there shouldn't be much situations were multiple are even possible or useful)
    res = grouped.agg(
        {
            "pretty_name": (lambda x: next(iter(set(x)))),
            "type_ids": (lambda x: next(iter(x))[0]),
            "types": (lambda x: next(iter(x))[0]),
            "row_nr": "nunique",
        }
    )
    res = res.rename(columns={"row_nr": "n_patient_visit"})
    # relative amount of patient visits with the specific entity to all patient visits (or rows in the original data)
    res["n_patient_visit_percent"] = (res["n_patient_visit"] / df["row_nr"].nunique()) * 100
    res.round({"n_patient_visit_percent": 1})

    return res


def _check_valid_name(df: pd.DataFrame, name: Iterable[str]) -> None:
    """Check whether the name is in the extracted entities to inform about possible typos.
    Currently, only the pretty_name column is supported.
    """
    invalid_names = []
    suggested_names = []

    for nm in name:
        pretty_names = df["pretty_name"].unique()

        if nm not in pretty_names:
            invalid_names.append(nm)
            try:
                new_name, _ = process.extractOne(query=nm, choices=pretty_names, score_cutoff=50)
                suggested_names.append(new_name)
            except EntitiyNotFoundError:
                pass

    if invalid_names:
        suggested_str = f" Do you mean {suggested_names}?" if suggested_names else ""
        msg = f"Did not find {invalid_names} in MedCAT's extracted entities and added them not to .obs.{suggested_str}"
        raise EntitiyNotFoundError(msg)


def add_medcat_annotation_to_obs(
    adata: AnnData,
    name: Iterable[str] | str,
    use_key: str = "medcat_annotations",
    added_colname: Iterable[str] | str | None = None,
    copy: bool = False,
) -> AnnData | None:
    """Add info extracted from free text as a binary column to obs.

    Indicates whether the specific entity to color by has been found in that row or not.


    Args:
        adata: AnnData object that holds the data to annotate.
        name: Name of the entity to add as a column to obs.
        use_key: Key to use for the annotated results.
        added_colname: Name of the column to add to obs. If None, name will be used.
        copy: Whether to copy adata or not.

    Returns:
        Returns `None` if `copy=False`, else returns an `AnnData` object. Sets the following fields;

        `adata.obs[name | added_coname]` : :class:`pandas.DataFrame`
            Added column(s) `to adata.obs`, indicating whether the specific entity to color by has been found in that row or not.

    """
    if use_key not in adata.uns.keys():
        raise ValueError(f"Key {use_key} not found in adata.uns. Please run ep.tl.annotate_text first.")

    if copy:
        adata = adata.copy()

    if isinstance(name, str):
        annotation_names = [name]
    else:
        annotation_names = list(name)

    if added_colname is None:
        added_colname = annotation_names
    elif isinstance(added_colname, str):
        added_colname = [added_colname]

    added_colnames = list(added_colname)
    if len(added_colnames) != len(annotation_names):
        raise ValueError(
            f"Length of added_colname ({len(added_colnames)}) does not match length of name ({len(annotation_names)})."
        )

    _check_valid_name(adata.uns[use_key], annotation_names)

    # only extract affirmed entities
    df = _filter_df_by_status(adata.uns[use_key], "Affirmed")

    # check whether the name is in the extracted entities to inform about possible typos
    # currently, only the pretty_name column is supported
    for i, annotation_name in enumerate(annotation_names):
        adata.obs[added_colnames[i]] = df.groupby("row_nr").agg(
            {
                "pretty_name": (
                    lambda row_pretty_names, annotation_name=annotation_name: any(
                        row_pretty_names.isin([annotation_name])
                    )
                )
            }
        )

    return adata if copy else None


class StatusNotSupportedError(Exception):
    pass


class EntitiyNotFoundError(Exception):
    pass



import os
import warnings
from pathlib import Path

import numpy as np
import pytest
from sklearn.exceptions import ConvergenceWarning

from ehrapy.io._read import read_csv
from ehrapy.preprocessing._imputation import (
    _warn_imputation_threshold,
    explicit_impute,
    iterative_svd_impute,
    knn_impute,
    matrix_factorization_impute,
    mice_forest_impute,
    miss_forest_impute,
    nuclear_norm_minimization_impute,
    simple_impute,
    soft_impute,
)

CURRENT_DIR = Path(__file__).parent
_TEST_PATH = f"{CURRENT_DIR}/test_data_imputation"


@pytest.fixture
def impute_num_adata():
    return read_csv(dataset_path=f"{_TEST_PATH}/test_impute_num.csv")


@pytest.fixture
def impute_adata():
    return read_csv(dataset_path=f"{_TEST_PATH}/test_impute.csv")


@pytest.fixture
def impute_iris():
    return read_csv(dataset_path=f"{_TEST_PATH}/test_impute_iris.csv")


@pytest.fixture
def impute_titanic():
    return read_csv(dataset_path=f"{_TEST_PATH}/test_impute_titanic.csv")


def test_mean_impute_no_copy(impute_num_adata):
    simple_impute(impute_num_adata)

    assert not np.isnan(impute_num_adata.X).any()


def test_mean_impute_copy(impute_num_adata):
    adata_imputed = simple_impute(impute_num_adata, copy=True)

    assert id(impute_num_adata) != id(adata_imputed)
    assert not np.isnan(adata_imputed.X).any()


def test_mean_impute_throws_error_non_numerical(impute_adata):
    with pytest.raises(ValueError):
        simple_impute(impute_adata)


def test_mean_impute_subset(impute_adata):
    adata_imputed = simple_impute(impute_adata, var_names=["intcol", "indexcol"], copy=True)

    assert not np.all([item != item for item in adata_imputed.X[::, 1:2]])
    assert np.any([item != item for item in adata_imputed.X[::, 3:4]])


def test_median_impute_no_copy(impute_num_adata):
    simple_impute(impute_num_adata, strategy="median")

    assert not np.isnan(impute_num_adata.X).any()


def test_median_impute_copy(impute_num_adata):
    adata_imputed = simple_impute(impute_num_adata, strategy="median", copy=True)

    assert id(impute_adata) != id(adata_imputed)
    assert not np.isnan(adata_imputed.X).any()


def test_median_impute_throws_error_non_numerical(impute_adata):
    with pytest.raises(ValueError):
        simple_impute(impute_adata, strategy="median")


def test_median_impute_subset(impute_adata):
    adata_imputed = simple_impute(impute_adata, var_names=["intcol", "indexcol"], strategy="median", copy=True)

    assert not np.all([item != item for item in adata_imputed.X[::, 1:2]])
    assert np.any([item != item for item in adata_imputed.X[::, 3:4]])


def test_most_frequent_impute_no_copy(impute_adata):
    simple_impute(impute_adata, strategy="most_frequent")

    assert not (np.all([item != item for item in impute_adata.X]))


def test_most_frequent_impute_copy(impute_adata):
    adata_imputed = simple_impute(impute_adata, strategy="most_frequent", copy=True)

    assert id(impute_adata) != id(adata_imputed)
    assert not (np.all([item != item for item in adata_imputed.X]))


def test_most_frequent_impute_subset(impute_adata):
    adata_imputed = simple_impute(impute_adata, var_names=["intcol", "strcol"], strategy="most_frequent", copy=True)

    assert not (np.all([item != item for item in adata_imputed.X[::, 1:3]]))


def test_knn_impute_no_copy(impute_num_adata):
    knn_impute(impute_num_adata)

    assert not (np.all([item != item for item in impute_num_adata.X]))


def test_knn_impute_copy(impute_num_adata):
    adata_imputed = knn_impute(impute_num_adata, n_neighbours=3, copy=True)

    assert id(impute_num_adata) != id(adata_imputed)
    assert not (np.all([item != item for item in adata_imputed.X]))


def test_knn_impute_non_numerical_data(impute_adata):
    adata_imputed = knn_impute(impute_adata, n_neighbours=3, copy=True)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_knn_impute_numerical_data(impute_num_adata):
    adata_imputed = knn_impute(impute_num_adata, copy=True)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_knn_impute_list_str(impute_adata):
    adata_imputed = knn_impute(impute_adata, var_names=["intcol", "strcol", "boolcol"], copy=True)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_missforest_impute_non_numerical_data(impute_adata):
    adata_imputed = miss_forest_impute(impute_adata, copy=True)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_missforest_impute_numerical_data(impute_num_adata):
    warnings.filterwarnings("ignore", category=ConvergenceWarning)
    adata_imputed = miss_forest_impute(impute_num_adata, copy=True)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_missforest_impute_subset(impute_num_adata):
    adata_imputed = miss_forest_impute(
        impute_num_adata, var_names={"non_numerical": ["intcol"], "numerical": ["strcol"]}, copy=True
    )

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_missforest_impute_list_str(impute_num_adata):
    warnings.filterwarnings("ignore", category=ConvergenceWarning)
    adata_imputed = miss_forest_impute(impute_num_adata, var_names=["col1", "col2", "col3"], copy=True)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_missforest_impute_dict(impute_adata):
    warnings.filterwarnings("ignore", category=ConvergenceWarning)
    adata_imputed = miss_forest_impute(
        impute_adata, var_names={"numerical": ["intcol", "datetime"], "non_numerical": ["strcol", "boolcol"]}, copy=True
    )

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_soft_impute_no_copy(impute_num_adata):
    adata_imputed = soft_impute(impute_num_adata)

    assert id(impute_num_adata) == id(adata_imputed)


def test_soft_impute_copy(impute_num_adata):
    adata_imputed = soft_impute(impute_num_adata, copy=True)

    assert id(impute_num_adata) != id(adata_imputed)


def test_soft_impute_non_numerical_data(impute_adata):
    adata_imputed = soft_impute(impute_adata)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_soft_impute_numerical_data(impute_num_adata):
    adata_imputed = soft_impute(impute_num_adata)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_soft_impute_list_str(impute_adata):
    adata_imputed = soft_impute(impute_adata, var_names=["intcol", "strcol", "boolcol"])

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_IterativeSVD_impute_no_copy(impute_num_adata):
    adata_imputed = iterative_svd_impute(impute_num_adata, rank=2)

    assert id(impute_num_adata) == id(adata_imputed)


def test_IterativeSVD_impute_copy(impute_num_adata):
    adata_imputed = iterative_svd_impute(impute_num_adata, rank=2, copy=True)

    assert id(impute_adata) != id(adata_imputed)


def test_IterativeSVD_impute_non_numerical_data(impute_adata):
    adata_imputed = iterative_svd_impute(impute_adata, rank=3)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_IterativeSVD_impute_numerical_data(impute_num_adata):
    adata_imputed = iterative_svd_impute(impute_num_adata, rank=2)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_IterativeSVD_impute_list_str(impute_adata):
    adata_imputed = iterative_svd_impute(impute_adata, var_names=["intcol", "strcol", "boolcol"], rank=2)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_matrix_factorization_impute_no_copy(impute_num_adata):
    adata_imputed = matrix_factorization_impute(impute_num_adata)

    assert id(impute_num_adata) == id(adata_imputed)


def test_matrix_factorization_impute_copy(impute_num_adata):
    adata_imputed = matrix_factorization_impute(impute_num_adata, copy=True)

    assert id(impute_num_adata) != id(adata_imputed)


def test_matrix_factorization_impute_non_numerical_data(impute_adata):
    adata_imputed = matrix_factorization_impute(impute_adata)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_matrix_factorization_impute_numerical_data(impute_adata):
    adata_imputed = matrix_factorization_impute(impute_adata)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_matrix_factorization_impute_list_str(impute_adata):
    adata_imputed = matrix_factorization_impute(impute_adata, var_names=["intcol", "strcol", "boolcol"])

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_nuclear_norm_minimization_impute_no_copy(impute_num_adata):
    adata_imputed = nuclear_norm_minimization_impute(impute_num_adata)

    assert id(impute_num_adata) == id(adata_imputed)


def test_nuclear_norm_minimization_impute_copy(impute_num_adata):
    adata_imputed = nuclear_norm_minimization_impute(impute_num_adata, copy=True)

    assert id(impute_num_adata) != id(adata_imputed)


def test_nuclear_norm_minimization_impute_non_numerical_data(impute_adata):
    adata_imputed = nuclear_norm_minimization_impute(impute_adata)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_nuclear_norm_minimization_impute_numerical_data(impute_num_adata):
    adata_imputed = nuclear_norm_minimization_impute(impute_num_adata)

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_nuclear_norm_minimization_impute_list_str(impute_adata):
    adata_imputed = nuclear_norm_minimization_impute(impute_adata, var_names=["intcol", "strcol", "boolcol"])

    assert not (np.all([item != item for item in adata_imputed.X]))


@pytest.mark.skipif(os.name == "posix", reason="miceforest Imputation not supported by MacOS.")
def test_miceforest_impute_no_copy(impute_iris):
    adata_imputed = mice_forest_impute(impute_iris)

    assert id(impute_iris) == id(adata_imputed)


@pytest.mark.skipif(os.name == "posix", reason="miceforest Imputation not supported by MacOS.")
def test_miceforest_impute_copy(impute_iris):
    adata_imputed = mice_forest_impute(impute_iris, copy=True)

    assert id(impute_iris) != id(adata_imputed)


@pytest.mark.skipif(os.name == "posix", reason="miceforest Imputation not supported by MacOS.")
def test_miceforest_impute_non_numerical_data(impute_titanic):
    adata_imputed = mice_forest_impute(impute_titanic)

    assert not (np.all([item != item for item in adata_imputed.X]))


@pytest.mark.skipif(os.name == "posix", reason="miceforest Imputation not supported by MacOS.")
def test_miceforest_impute_numerical_data(impute_iris):
    adata_imputed = mice_forest_impute(impute_iris)

    assert not (np.all([item != item for item in adata_imputed.X]))


@pytest.mark.skipif(os.name == "posix", reason="miceforest Imputation not supported by MacOS.")
def test_miceforest_impute_list_str(impute_titanic):
    adata_imputed = mice_forest_impute(impute_titanic, var_names=["Cabin", "Age"])

    assert not (np.all([item != item for item in adata_imputed.X]))


def test_explicit_impute_all(impute_num_adata):
    warnings.filterwarnings("ignore", category=FutureWarning)
    adata_imputed = explicit_impute(impute_num_adata, replacement=1011, copy=True)

    assert (adata_imputed.X == 1011).sum() == 3


def test_explicit_impute_subset(impute_adata):
    adata_imputed = explicit_impute(impute_adata, replacement={"strcol": "REPLACED", "intcol": 1011}, copy=True)

    assert (adata_imputed.X == 1011).sum() == 1
    assert (adata_imputed.X == "REPLACED").sum() == 1


def test_warning(impute_num_adata):
    warning_results = _warn_imputation_threshold(impute_num_adata, threshold=20, var_names=None)
    assert warning_results == {"col1": 25, "col3": 50}
